from __future__ import annotations

from collections.abc import Callable

import pytest

from src.microsoft_mcp.tools.calendar import calendar_delete_event
from src.microsoft_mcp.tools.contact import contact_delete
from src.microsoft_mcp.tools.email import email_delete, email_reply, email_send
from src.microsoft_mcp.tools.email_rules import emailrules_delete
from src.microsoft_mcp.tools.file import file_delete
from src.microsoft_mcp.validators import ValidationError


def _record(response: object, calls: list[str]) -> Callable[[], object]:
    """Return a callable that records when Graph is invoked."""

    def _wrapped() -> object:
        calls.append("called")
        return response

    return _wrapped


def test_email_delete_requires_confirmation(
    mock_graph_request,
    mock_account_id: str,
) -> None:
    email_id = "mock-email-id"

    with pytest.raises(ValidationError):
        email_delete.fn(email_id, mock_account_id, confirm=False)

    calls: list[str] = []
    mock_graph_request(
        "DELETE",
        f"/me/messages/{email_id}",
        _record(None, calls),
    )

    result = email_delete.fn(email_id, mock_account_id, confirm=True)
    assert result == {"status": "deleted"}
    assert calls, "Graph request should execute when confirm=True"


def test_file_delete_requires_confirmation(
    mock_graph_request,
    mock_account_id: str,
) -> None:
    file_id = "01ABCDEFZLMNO!123"

    with pytest.raises(ValidationError):
        file_delete.fn(file_id, mock_account_id, confirm=False)

    calls: list[str] = []
    mock_graph_request(
        "DELETE",
        f"/me/drive/items/{file_id}",
        _record(None, calls),
    )

    result = file_delete.fn(file_id, mock_account_id, confirm=True)
    assert result == {"status": "deleted"}
    assert calls, "Graph request should execute when confirm=True"


def test_contact_delete_requires_confirmation(
    mock_graph_request,
    mock_account_id: str,
) -> None:
    contact_id = "contact-123"

    with pytest.raises(ValidationError):
        contact_delete.fn(contact_id, mock_account_id, confirm=False)

    calls: list[str] = []
    mock_graph_request(
        "DELETE",
        f"/me/contacts/{contact_id}",
        _record(None, calls),
    )

    result = contact_delete.fn(contact_id, mock_account_id, confirm=True)
    assert result == {"status": "deleted"}
    assert calls, "Graph request should execute when confirm=True"


def test_calendar_delete_event_requires_confirmation(
    mock_graph_request,
    mock_account_id: str,
) -> None:
    event_id = "event-123"

    with pytest.raises(ValidationError):
        calendar_delete_event.fn(
            mock_account_id,
            event_id,
            send_cancellation=False,
            confirm=False,
        )

    calls: list[str] = []
    mock_graph_request(
        "DELETE",
        f"/me/events/{event_id}",
        _record(None, calls),
    )

    result = calendar_delete_event.fn(
        mock_account_id,
        event_id,
        send_cancellation=False,
        confirm=True,
    )
    assert result == {"status": "deleted"}
    assert calls, "Graph request should execute when confirm=True"


def test_emailrules_delete_requires_confirmation(
    mock_graph_request,
    mock_account_id: str,
) -> None:
    rule_id = "rule-123"

    with pytest.raises(ValidationError):
        emailrules_delete.fn(rule_id, mock_account_id, confirm=False)

    calls: list[str] = []
    mock_graph_request(
        "DELETE",
        f"/me/mailFolders/inbox/messageRules/{rule_id}",
        _record(None, calls),
    )

    result = emailrules_delete.fn(rule_id, mock_account_id, confirm=True)
    assert result == {"status": "deleted", "rule_id": rule_id}
    assert calls, "Graph request should execute when confirm=True"


def test_email_send_requires_confirmation(
    mock_graph_request,
    mock_account_id: str,
) -> None:
    with pytest.raises(ValidationError):
        email_send.fn(
            account_id=mock_account_id,
            to="recipient@example.com",
            subject="Test",
            body="Hello",
            confirm=False,
        )

    calls: list[str] = []
    mock_graph_request(
        "POST",
        "/me/sendMail",
        _record({"status": "sent"}, calls),
    )

    result = email_send.fn(
        account_id=mock_account_id,
        to="recipient@example.com",
        subject="Test",
        body="Hello",
        confirm=True,
    )
    assert result == {"status": "sent"}
    assert calls, "Graph request should execute when confirm=True"


def test_email_reply_requires_confirmation(
    mock_graph_request,
    mock_account_id: str,
) -> None:
    email_id = "mock-email-id"

    with pytest.raises(ValidationError):
        email_reply.fn(
            account_id=mock_account_id,
            email_id=email_id,
            body="Hello",
            confirm=False,
        )

    calls: list[str] = []
    mock_graph_request(
        "POST",
        f"/me/messages/{email_id}/reply",
        _record(None, calls),
    )

    result = email_reply.fn(
        account_id=mock_account_id,
        email_id=email_id,
        body="Hello",
        confirm=True,
    )
    assert result == {"status": "sent"}
    assert calls, "Graph request should execute when confirm=True"



"""Shared parameter validation helpers for Microsoft MCP tools."""

from __future__ import annotations

import logging
import os
import re
import tempfile
from datetime import datetime
from pathlib import Path
from typing import Any, Collection, Iterable, Sequence
from urllib.parse import urljoin, urlparse

from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

LOGGER = logging.getLogger("microsoft_mcp.validators")

EMAIL_PATTERN = re.compile(
    r"^(?P<local>[A-Za-z0-9.!#$%&'*+/=?^_`{|}~-]+)"
    r"@(?P<domain>[A-Za-z0-9-]+(?:\.[A-Za-z0-9-]+)*)$"
)

GRAPH_ALLOWED_HOSTS = {
    "graph.microsoft.com",
    "graph.microsoft.us",
    "graph.microsoft.de",
    "microsoftgraph.chinacloudapi.cn",
}

GRAPH_ALLOWED_SUFFIXES = (
    ".sharepoint.com",
    ".sharepoint-df.com",
    ".sharepoint-us.com",
    ".onmicrosoft.com",
    ".1drv.com",
    ".onedrive.live.com",
    ".storage.live.com",
    ".microsoftpersonalcontent.com",
)

WINDOWS_RESERVED_NAMES = {
    "CON",
    "PRN",
    "AUX",
    "NUL",
    *(f"COM{i}" for i in range(1, 10)),
    *(f"LPT{i}" for i in range(1, 10)),
}

DEFAULT_ALLOWED_ROOTS: tuple[Path, ...] = (
    Path.cwd().resolve(),
    Path(tempfile.gettempdir()).resolve(),
)


class ValidationError(ValueError):
    """Raised when parameter validation fails."""


def _mask_value(value: Any) -> str:
    """Return a sanitised representation of a potentially sensitive value."""
    if value is None:
        return "None"

    if isinstance(value, str):
        stripped = value.strip()
        if not stripped:
            return ""

        email_match = EMAIL_PATTERN.match(stripped)
        if email_match:
            local = email_match.group("local")
            domain = email_match.group("domain")
            if len(local) <= 1:
                masked_local = "*"
            elif len(local) == 2:
                masked_local = local[0] + "*"
            else:
                masked_local = local[0] + "***" + local[-1]
            return f"{masked_local}@{domain}"

        if len(stripped) > 64:
            return f"{stripped[:32]}…{stripped[-8:]}"

        return stripped

    return str(value)


def format_validation_error(
    param: str,
    value: Any,
    reason: str,
    expected: str,
) -> str:
    """Format the canonical validation error message."""
    masked = _mask_value(value)
    return f"Invalid {param} '{masked}': {reason}. Expected: {expected}"


def _log_failure(param: str, reason: str, value: Any) -> None:
    """Log validation failure without exposing sensitive data."""
    LOGGER.warning(
        "Validation failed",
        extra={
            "param": param,
            "reason": reason,
            "value": _mask_value(value),
        },
    )


def validate_account_id(account_id: str, param_name: str = "account_id") -> str:
    """Ensure account identifiers are present and well formed."""
    if not isinstance(account_id, str):
        reason = "must be a string"
        _log_failure(param_name, reason, account_id)
        raise ValidationError(
            format_validation_error(param_name, account_id, reason, "non-empty string")
        )

    trimmed = account_id.strip()
    if not trimmed:
        reason = "cannot be empty"
        _log_failure(param_name, reason, account_id)
        raise ValidationError(
            format_validation_error(param_name, account_id, reason, "non-empty string")
        )

    return trimmed


def validate_confirmation_flag(
    confirm: bool | None,
    operation: str,
    resource_type: str,
    param_name: str = "confirm",
) -> bool:
    """Ensure destructive operations require explicit confirmation."""
    if confirm is not True:
        reason = f"{operation} on {resource_type} requires confirm=True to proceed"
        _log_failure(param_name, reason, confirm)
        raise ValidationError(
            format_validation_error(
                param_name,
                confirm,
                reason,
                "Explicit user confirmation",
            )
        )
    return True


def require_confirm(confirm: bool, action: str, param_name: str = "confirm") -> None:
    """Backward compatible confirmation guard."""
    validate_confirmation_flag(confirm, action, "resource", param_name)


def validate_positive_int(value: Any, name: str) -> int:
    """Return a positive integer or raise ValidationError."""
    if not isinstance(value, int):
        reason = "must be an integer"
        _log_failure(name, reason, value)
        raise ValidationError(
            format_validation_error(name, value, reason, "Positive integer")
        )
    if value <= 0:
        reason = "must be greater than zero"
        _log_failure(name, reason, value)
        raise ValidationError(
            format_validation_error(name, value, reason, "Positive integer (> 0)")
        )
    return value


def validate_limit(
    limit: Any,
    minimum: int,
    maximum: int,
    param_name: str = "limit",
) -> int:
    """Validate pagination or item limits."""
    if not isinstance(limit, int):
        reason = "must be an integer"
        _log_failure(param_name, reason, limit)
        raise ValidationError(
            format_validation_error(param_name, limit, reason, f"{minimum}-{maximum}")
        )
    if limit < minimum or limit > maximum:
        reason = f"must be between {minimum} and {maximum}"
        _log_failure(param_name, reason, limit)
        raise ValidationError(
            format_validation_error(param_name, limit, reason, f"{minimum}-{maximum}")
        )
    return limit


def validate_email_format(email: str, param_name: str = "email") -> str:
    """Validate email address format and normalise casing."""
    if not isinstance(email, str):
        reason = "must be a string"
        _log_failure(param_name, reason, email)
        raise ValidationError(
            format_validation_error(param_name, email, reason, "Valid email address")
        )
    trimmed = email.strip()
    match = EMAIL_PATTERN.match(trimmed)
    if not match:
        reason = "does not match email format"
        _log_failure(param_name, reason, email)
        raise ValidationError(
            format_validation_error(
                param_name, email, reason, "name@example.com pattern"
            )
        )
    return trimmed.lower()


def normalize_recipients(
    recipients: str | Sequence[str] | None,
    param_name: str = "recipients",
) -> list[str]:
    """Normalise recipient inputs to a list of validated email addresses."""
    if recipients is None:
        return []
    values: Iterable[str]
    if isinstance(recipients, str):
        values = [part.strip() for part in recipients.split(",")]
    else:
        values = recipients

    normalised: list[str] = []
    for raw in values:
        if not raw:
            continue
        normalised.append(validate_email_format(raw, param_name))

    if not normalised:
        reason = "must include at least one valid email"
        _log_failure(param_name, reason, recipients)
        raise ValidationError(
            format_validation_error(
                param_name,
                recipients,
                reason,
                "Non-empty recipient list",
            )
        )
    return normalised


def validate_choices(
    value: str,
    allowed: Collection[str],
    param_name: str = "value",
) -> str:
    """Validate that a string value is within an allowed set (case-insensitive)."""
    if not isinstance(value, str):
        reason = "must be a string"
        _log_failure(param_name, reason, value)
        raise ValidationError(
            format_validation_error(
                param_name, value, reason, f"One of {sorted(allowed)}"
            )
        )

    trimmed = value.strip()
    if not trimmed:
        reason = "cannot be empty"
        _log_failure(param_name, reason, value)
        raise ValidationError(
            format_validation_error(
                param_name, value, reason, f"One of {sorted(allowed)}"
            )
        )

    allowed_map = {item.casefold(): item for item in allowed}
    matched = allowed_map.get(trimmed.casefold())
    if matched is None:
        reason = "not in allowed set"
        _log_failure(param_name, reason, value)
        raise ValidationError(
            format_validation_error(
                param_name, value, reason, f"One of {sorted(allowed)}"
            )
        )
    return matched


def validate_iso_datetime(
    value: str,
    param_name: str,
    allow_date_only: bool = False,
) -> datetime:
    """Parse ISO-8601 datetimes ensuring timezone awareness."""
    if not isinstance(value, str):
        reason = "must be a string"
        _log_failure(param_name, reason, value)
        raise ValidationError(
            format_validation_error(param_name, value, reason, "ISO-8601 string")
        )

    trimmed = value.strip()
    if not trimmed:
        reason = "cannot be empty"
        _log_failure(param_name, reason, value)
        raise ValidationError(
            format_validation_error(param_name, value, reason, "ISO-8601 string")
        )

    try:
        parsed = datetime.fromisoformat(trimmed.replace("Z", "+00:00"))
    except ValueError as exc:
        reason = f"invalid ISO-8601 datetime ({exc})"
        _log_failure(param_name, reason, value)
        raise ValidationError(
            format_validation_error(param_name, value, reason, "YYYY-MM-DDTHH:MM:SS+TZ")
        ) from exc

    if parsed.tzinfo is None:
        if allow_date_only and len(trimmed) == 10:
            return parsed
        reason = "timezone information required"
        _log_failure(param_name, reason, value)
        raise ValidationError(
            format_validation_error(
                param_name, value, reason, "Timezone-aware datetime"
            )
        )

    return parsed


def validate_timezone(tz: str, param_name: str = "timezone") -> str:
    """Ensure timezone names resolve via zoneinfo."""
    if not isinstance(tz, str):
        reason = "must be a string"
        _log_failure(param_name, reason, tz)
        raise ValidationError(
            format_validation_error(param_name, tz, reason, "Valid IANA timezone name")
        )
    trimmed = tz.strip()
    if not trimmed:
        reason = "cannot be empty"
        _log_failure(param_name, reason, tz)
        raise ValidationError(
            format_validation_error(param_name, tz, reason, "Valid IANA timezone name")
        )
    try:
        ZoneInfo(trimmed)
    except ZoneInfoNotFoundError as exc:
        reason = "unknown timezone"
        _log_failure(param_name, reason, tz)
        raise ValidationError(
            format_validation_error(param_name, tz, reason, "Valid IANA timezone name")
        ) from exc
    return trimmed


def validate_datetime_window(
    start: str,
    end: str,
    *,
    allow_equal: bool = False,
) -> tuple[datetime, datetime]:
    """Validate a datetime range."""
    start_dt = validate_iso_datetime(start, "start")
    end_dt = validate_iso_datetime(end, "end")

    if allow_equal and start_dt == end_dt:
        return start_dt, end_dt

    if start_dt >= end_dt:
        reason = "start must be earlier than end"
        _log_failure("datetime_window", reason, f"{start_dt}->{end_dt}")
        raise ValidationError(
            format_validation_error(
                "datetime_window",
                f"{_mask_value(start)}→{_mask_value(end)}",
                reason,
                "start < end",
            )
        )
    return start_dt, end_dt


def validate_datetime_ordering(
    first: datetime,
    second: datetime,
    param_name: str = "datetime_order",
) -> None:
    """Ensure the first datetime is not after the second."""
    if first > second:
        reason = "first datetime must not be after second"
        _log_failure(param_name, reason, f"{first}->{second}")
        raise ValidationError(
            format_validation_error(
                param_name,
                f"{first.isoformat()}→{second.isoformat()}",
                reason,
                "First <= Second",
            )
        )


def validate_folder_choice(
    value: str,
    allowed: Sequence[str],
    param_name: str = "folder",
) -> str:
    """Validate folder selections against allowed list."""
    if not isinstance(value, str):
        reason = "must be a string"
        _log_failure(param_name, reason, value)
        raise ValidationError(
            format_validation_error(param_name, value, reason, f"One of {allowed}")
        )
    normalised = value.strip()
    allowed_lower = {item.lower(): item for item in allowed}
    if normalised.lower() not in allowed_lower:
        reason = "not in allowed set"
        _log_failure(param_name, reason, value)
        raise ValidationError(
            format_validation_error(
                param_name, value, reason, f"One of {sorted(allowed)}"
            )
        )
    return allowed_lower[normalised.lower()]


def validate_json_payload(
    payload: Any,
    *,
    required_keys: Sequence[str] | None = None,
    allowed_keys: Sequence[str] | None = None,
    param_name: str = "payload",
) -> dict[str, Any]:
    """Ensure payloads are JSON-like dictionaries with required/allowed keys."""
    if not isinstance(payload, dict):
        reason = "must be a JSON object"
        _log_failure(param_name, reason, payload)
        raise ValidationError(
            format_validation_error(param_name, payload, reason, "Dictionary value")
        )

    if required_keys:
        missing = [key for key in required_keys if key not in payload]
        if missing:
            reason = f"missing keys {missing}"
            _log_failure(param_name, reason, payload)
            raise ValidationError(
                format_validation_error(
                    param_name,
                    payload,
                    reason,
                    f"Include keys {sorted(required_keys)}",
                )
            )

    if allowed_keys is not None:
        allowed_set = set(allowed_keys)
        unknown = sorted(key for key in payload if key not in allowed_set)
        if unknown:
            reason = f"contains unsupported keys {unknown}"
            _log_failure(param_name, reason, payload)
            raise ValidationError(
                format_validation_error(
                    param_name,
                    payload,
                    reason,
                    f"Only keys {sorted(allowed_set)}",
                )
            )

    return payload


def validate_request_size(
    size_bytes: int,
    limit_bytes: int,
    param_name: str = "request_size",
) -> int:
    """Ensure request payloads stay within configured limits."""
    if not isinstance(size_bytes, int):
        reason = "must be integer number of bytes"
        _log_failure(param_name, reason, size_bytes)
        raise ValidationError(
            format_validation_error(
                param_name, size_bytes, reason, "Integer bytes value"
            )
        )
    if size_bytes < 0:
        reason = "cannot be negative"
        _log_failure(param_name, reason, size_bytes)
        raise ValidationError(
            format_validation_error(
                param_name, size_bytes, reason, "0 or positive integer"
            )
        )
    if size_bytes > limit_bytes:
        reason = f"exceeds limit of {limit_bytes} bytes"
        _log_failure(param_name, reason, size_bytes)
        raise ValidationError(
            format_validation_error(
                param_name,
                size_bytes,
                reason,
                f"<= {limit_bytes} bytes",
            )
        )
    return size_bytes


def validate_microsoft_graph_id(
    identifier: str,
    param_name: str = "identifier",
) -> str:
    """Validate Microsoft Graph resource identifiers."""
    if not isinstance(identifier, str):
        reason = "must be a string"
        _log_failure(param_name, reason, identifier)
        raise ValidationError(
            format_validation_error(
                param_name,
                identifier,
                reason,
                "Graph resource identifier string",
            )
        )
    trimmed = identifier.strip()
    if not trimmed:
        reason = "cannot be empty"
        _log_failure(param_name, reason, identifier)
        raise ValidationError(
            format_validation_error(param_name, identifier, reason, "Non-empty string")
        )
    if not re.fullmatch(r"[A-Za-z0-9\-._!]{1,256}", trimmed):
        reason = "contains unsupported characters"
        _log_failure(param_name, reason, identifier)
        raise ValidationError(
            format_validation_error(
                param_name,
                identifier,
                reason,
                "Alphanumeric with - . _ !",
            )
        )
    return trimmed


def validate_onedrive_path(
    onedrive_path: str,
    param_name: str = "onedrive_path",
) -> str:
    """Validate OneDrive path format."""
    if not isinstance(onedrive_path, str):
        reason = "must be a string"
        _log_failure(param_name, reason, onedrive_path)
        raise ValidationError(
            format_validation_error(
                param_name,
                onedrive_path,
                reason,
                "Path beginning with '/'",
            )
        )
    trimmed = onedrive_path.strip()
    if not trimmed.startswith("/"):
        reason = "must start with '/'"
        _log_failure(param_name, reason, onedrive_path)
        raise ValidationError(
            format_validation_error(
                param_name,
                onedrive_path,
                reason,
                "Absolute OneDrive path like '/Documents/file.txt'",
            )
        )
    parts = [part for part in trimmed.split("/") if part]
    invalid_chars = re.compile(r'[<>:"|?*]')
    for part in parts:
        if part == "..":
            reason = "parent directory segments are not allowed"
            _log_failure(param_name, reason, onedrive_path)
            raise ValidationError(
                format_validation_error(
                    param_name,
                    onedrive_path,
                    reason,
                    "Path without '..' segments",
                )
            )
        if invalid_chars.search(part):
            reason = "contains reserved characters"
            _log_failure(param_name, reason, onedrive_path)
            raise ValidationError(
                format_validation_error(
                    param_name,
                    onedrive_path,
                    reason,
                    'Path without < > : " | ? * characters',
                )
            )
        if os.name == "nt":
            upper = part.split(".")[0].upper()
            if upper in WINDOWS_RESERVED_NAMES:
                reason = "contains Windows reserved filename"
                _log_failure(param_name, reason, onedrive_path)
                raise ValidationError(
                    format_validation_error(
                        param_name,
                        onedrive_path,
                        reason,
                        "Path without reserved Windows names",
                    )
                )
    normalised = "/" + "/".join(parts)
    return normalised or "/"


def _resolve_allowed_roots(
    allowed_roots: Sequence[str | Path] | None,
) -> list[Path]:
    """Resolve allowed root directories."""
    roots: list[Path] = []
    if allowed_roots:
        for entry in allowed_roots:
            roots.append(Path(entry).expanduser().resolve())
    else:
        roots.extend(DEFAULT_ALLOWED_ROOTS)

    env_roots = os.getenv("MCP_FILE_ALLOWED_ROOTS")
    if env_roots:
        for part in env_roots.split(os.pathsep):
            if part:
                roots.append(Path(part).expanduser().resolve())
    return roots


def _is_subpath(candidate: Path, parent: Path) -> bool:
    """Return True if candidate resides within parent."""
    try:
        candidate.relative_to(parent)
        return True
    except ValueError:
        return False


def ensure_safe_path(
    path: str | Path,
    *,
    allow_overwrite: bool = False,
    must_exist: bool = False,
    allowed_roots: Sequence[str | Path] | None = None,
    param_name: str = "path",
) -> Path:
    """Validate local filesystem paths for safe read/write operations."""
    candidate = Path(path).expanduser()
    candidate_str = str(candidate)

    if os.name == "nt":
        if candidate_str.startswith("\\\\"):
            reason = "UNC paths are not allowed"
            _log_failure(param_name, reason, candidate_str)
            raise ValidationError(
                format_validation_error(
                    param_name,
                    candidate_str,
                    reason,
                    "Absolute path on local drive",
                )
            )

    resolved = candidate.resolve(strict=False)

    if ".." in candidate.parts:
        reason = "parent directory segments are not allowed"
        _log_failure(param_name, reason, candidate_str)
        raise ValidationError(
            format_validation_error(
                param_name,
                candidate_str,
                reason,
                "Path without '..' segments",
            )
        )

    roots = _resolve_allowed_roots(allowed_roots)
    if not any(_is_subpath(resolved, root) for root in roots):
        reason = "path escapes allowed directories"
        _log_failure(param_name, reason, candidate_str)
        raise ValidationError(
            format_validation_error(
                param_name,
                candidate_str,
                reason,
                f"Location inside {', '.join(str(r) for r in roots)}",
            )
        )

    if os.name == "nt":
        for part in resolved.parts:
            if part == resolved.drive:
                continue
            if ":" in part:
                reason = "alternate data streams are not allowed"
                _log_failure(param_name, reason, candidate_str)
                raise ValidationError(
                    format_validation_error(
                        param_name,
                        candidate_str,
                        reason,
                        "Filenames without ':' segments",
                    )
                )
            upper = part.split(".")[0].upper()
            if upper in WINDOWS_RESERVED_NAMES:
                reason = "contains Windows reserved filename"
                _log_failure(param_name, reason, candidate_str)
                raise ValidationError(
                    format_validation_error(
                        param_name,
                        candidate_str,
                        reason,
                        "No reserved Windows filenames",
                    )
                )

    if resolved.exists():
        if resolved.is_dir():
            reason = "expected file path but found directory"
            _log_failure(param_name, reason, candidate_str)
            raise ValidationError(
                format_validation_error(
                    param_name,
                    candidate_str,
                    reason,
                    "Existing file or new file path",
                )
            )
        if not allow_overwrite:
            reason = "would overwrite existing file"
            _log_failure(param_name, reason, candidate_str)
            raise ValidationError(
                format_validation_error(
                    param_name,
                    candidate_str,
                    reason,
                    "Unique file path",
                )
            )
    elif must_exist:
        reason = "path does not exist"
        _log_failure(param_name, reason, candidate_str)
        raise ValidationError(
            format_validation_error(
                param_name,
                candidate_str,
                reason,
                "Existing file path",
            )
        )

    parent = resolved.parent
    if not parent.exists():
        try:
            parent_relative = parent.relative_to(parent.anchor)
            if any(part == ".." for part in parent_relative.parts):
                reason = "parent directory traversal detected"
                _log_failure(param_name, reason, candidate_str)
                raise ValidationError(
                    format_validation_error(
                        param_name,
                        candidate_str,
                        reason,
                        "Valid parent directory",
                    )
                )
        except ValueError:
            pass

    if resolved.is_symlink():
        reason = "symlinks are not allowed"
        _log_failure(param_name, reason, candidate_str)
        raise ValidationError(
            format_validation_error(
                param_name,
                candidate_str,
                reason,
                "Regular file path",
            )
        )

    return resolved


def validate_graph_url(
    url: str,
    param_name: str = "url",
    *,
    allow_redirect: bool = False,
) -> str:
    """Validate download URLs returned by Microsoft Graph."""
    if not isinstance(url, str):
        reason = "must be a string"
        _log_failure(param_name, reason, url)
        raise ValidationError(
            format_validation_error(param_name, url, reason, "HTTPS URL")
        )

    trimmed = url.strip()
    if not trimmed:
        reason = "cannot be empty"
        _log_failure(param_name, reason, url)
        raise ValidationError(
            format_validation_error(param_name, url, reason, "HTTPS URL")
        )

    parsed = urlparse(trimmed)
    if parsed.scheme.lower() != "https":
        reason = "must use HTTPS"
        _log_failure(param_name, reason, url)
        raise ValidationError(
            format_validation_error(param_name, url, reason, "HTTPS URL")
        )

    host = parsed.hostname or ""
    host_lower = host.lower()
    allowed = host_lower in GRAPH_ALLOWED_HOSTS or any(
        host_lower.endswith(suffix) for suffix in GRAPH_ALLOWED_SUFFIXES
    )
    if not allowed:
        reason = "host is not an approved Microsoft domain"
        _log_failure(param_name, reason, host_lower)
        raise ValidationError(
            format_validation_error(
                param_name,
                url,
                reason,
                f"Host within {sorted(GRAPH_ALLOWED_HOSTS)} or "
                f"{GRAPH_ALLOWED_SUFFIXES}",
            )
        )

    if parsed.username or parsed.password:
        reason = "embedded credentials are not allowed"
        _log_failure(param_name, reason, url)
        raise ValidationError(
            format_validation_error(
                param_name,
                url,
                reason,
                "URL without credentials",
            )
        )

    if allow_redirect and parsed.path:
        trimmed = urljoin(trimmed, parsed.path)

    return trimmed


def validate_attachments(
    attachments: Sequence[str | Path] | str | None,
    *,
    max_inline_size_bytes: int = 3_145_728,
    max_attachments: int = 10,
    param_name: str = "attachments",
) -> list[Path]:
    """Validate attachment inputs for email tools."""
    if attachments is None:
        return []

    if isinstance(attachments, (str, Path)):
        values = [attachments]
    else:
        values = list(attachments)

    if not values:
        reason = "attachment list cannot be empty"
        _log_failure(param_name, reason, attachments)
        raise ValidationError(
            format_validation_error(
                param_name,
                attachments,
                reason,
                "Provide at least one attachment or None",
            )
        )

    if len(values) > max_attachments:
        reason = f"exceeds limit of {max_attachments} attachments"
        _log_failure(param_name, reason, len(values))
        raise ValidationError(
            format_validation_error(
                param_name,
                len(values),
                reason,
                f"<= {max_attachments} attachments",
            )
        )

    validated: list[Path] = []
    for item in values:
        path = ensure_safe_path(
            Path(item),
            allow_overwrite=True,
            must_exist=True,
            param_name=param_name,
        )
        size = path.stat().st_size
        validate_request_size(size, max_inline_size_bytes, f"{param_name}_size")
        validated.append(path)
    return validated



from __future__ import annotations

import logging
import os
import time
from pathlib import Path
from typing import Any
from urllib.parse import urljoin

import httpx

from .. import graph
from ..mcp_instance import mcp
from ..validators import (
    ValidationError,
    ensure_safe_path,
    format_validation_error,
    require_confirm,
    validate_account_id,
    validate_graph_url,
    validate_limit,
    validate_microsoft_graph_id,
    validate_onedrive_path,
    validate_request_size,
)

LOGGER = logging.getLogger("microsoft_mcp.tools.file")

DEFAULT_DOWNLOAD_TIMEOUT = float(os.getenv("MCP_FILE_DOWNLOAD_TIMEOUT", "60.0"))
DEFAULT_CHUNK_SIZE = int(os.getenv("MCP_FILE_DOWNLOAD_CHUNK_SIZE", "1048576"))
MAX_DOWNLOAD_MIB = int(os.getenv("MCP_FILE_DOWNLOAD_MAX_MB", "512"))
MAX_REDIRECTS = 3


# file_list
@mcp.tool(
    name="file_list",
    annotations={
        "title": "List Files",
        "readOnlyHint": True,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={"category": "file", "safety_level": "safe"},
)
def file_list(
    account_id: str,
    path: str = "/",
    folder_id: str | None = None,
    limit: int = 50,
    type_filter: str = "all",
) -> list[dict[str, Any]]:
    """📖 List files and/or folders in OneDrive (read-only, safe for unsupervised use)

    Returns items from OneDrive with names, sizes, and modification dates.

    Args:
        account_id: Microsoft account ID
        path: Path to list from (default: "/")
        folder_id: Direct folder ID (takes precedence over path)
        limit: Maximum items to return (1-500, default: 50)
        type_filter: Filter by type - "all", "files", or "folders" (default: "all")

    Returns:
        List of items matching the filter criteria
    """
    validate_account_id(account_id)
    limit = validate_limit(limit, 1, 500)
    if type_filter not in ["all", "files", "folders"]:
        raise ValidationError(
            format_validation_error(
                "type_filter",
                type_filter,
                "unsupported value",
                "'all', 'files', or 'folders'",
            )
        )

    # Determine endpoint
    if folder_id:
        endpoint = f"/me/drive/items/{folder_id}/children"
    else:
        validated_path = validate_onedrive_path(path, "path")
        if validated_path == "/":
            endpoint = "/me/drive/root/children"
        else:
            endpoint = f"/me/drive/root:{validated_path}:/children"

    params = {
        "$top": limit,
        "$select": "id,name,size,lastModifiedDateTime,folder,file,@microsoft.graph.downloadUrl",
    }

    items = list(
        graph.request_paginated(endpoint, account_id, params=params, limit=limit)
    )

    # Apply type filtering
    result = []
    for item in items:
        is_folder = "folder" in item
        is_file = "file" in item

        # Filter based on type_filter
        if type_filter == "folders" and not is_folder:
            continue
        if type_filter == "files" and not is_file:
            continue

        result.append(
            {
                "id": item["id"],
                "name": item["name"],
                "type": "folder" if is_folder else "file",
                "size": item.get("size", 0),
                "modified": item.get("lastModifiedDateTime"),
                "download_url": item.get("@microsoft.graph.downloadUrl"),
            }
        )

    return result


def _stream_download(
    url: str,
    destination: Path,
    *,
    timeout: float,
    chunk_size: int,
) -> None:
    """Stream file contents from a validated URL to destination path."""
    target_url = url
    for redirect in range(MAX_REDIRECTS + 1):
        with httpx.Client(timeout=timeout, follow_redirects=False) as client:
            with client.stream("GET", target_url) as response:
                if response.status_code in {301, 302, 303, 307, 308}:
                    location = response.headers.get("Location")
                    if not location:
                        raise RuntimeError("Redirect response missing Location header")
                    next_url = urljoin(target_url, location)
                    target_url = validate_graph_url(next_url, "redirect_url")
                    continue

                response.raise_for_status()
                with destination.open("wb") as output:
                    for chunk in response.iter_bytes(chunk_size):
                        if chunk:
                            output.write(chunk)
                return
    raise RuntimeError("Exceeded redirect limit during download")


def _download_with_retries(
    url: str,
    destination: Path,
    *,
    timeout: float,
    chunk_size: int,
    retries: int,
) -> None:
    """Download with exponential backoff retry strategy."""
    attempt = 0
    backoff = 1.0
    last_error: Exception | None = None
    while attempt <= retries:
        try:
            _stream_download(
                url,
                destination,
                timeout=timeout,
                chunk_size=chunk_size,
            )
            return
        except (httpx.HTTPError, OSError, RuntimeError) as exc:
            last_error = exc
            LOGGER.warning(
                "file_get download attempt failed",
                extra={
                    "attempt": attempt + 1,
                    "retries": retries,
                    "reason": str(exc),
                },
            )
            if destination.exists():
                destination.unlink(missing_ok=True)
            if attempt >= retries:
                break
            time.sleep(backoff)
            backoff *= 2
            attempt += 1
    assert last_error is not None
    raise RuntimeError(f"Failed to download file after retries: {last_error}") from (
        last_error
    )


def _file_get_impl(file_id: str, account_id: str, download_path: str) -> dict[str, Any]:
    account = validate_account_id(account_id)
    graph_file_id = validate_microsoft_graph_id(file_id, "file_id")
    destination = ensure_safe_path(download_path, allow_overwrite=False)
    destination.parent.mkdir(parents=True, exist_ok=True)

    metadata = graph.request("GET", f"/me/drive/items/{graph_file_id}", account)
    if not metadata:
        raise ValidationError(
            format_validation_error(
                "file_id",
                graph_file_id,
                "not found in OneDrive",
                "Existing file identifier",
            )
        )

    size_bytes = metadata.get("size", 0) or 0
    size_limit = MAX_DOWNLOAD_MIB * 1024 * 1024
    validate_request_size(size_bytes, size_limit, "download_size")

    download_url = metadata.get("@microsoft.graph.downloadUrl")
    if not download_url:
        raise RuntimeError("No download URL available for this file")
    safe_url = validate_graph_url(download_url, "download_url")

    LOGGER.info(
        "Initiating file download",
        extra={
            "file_id": graph_file_id,
            "account_id": account,
            "destination": str(destination),
            "expected_size": size_bytes,
        },
    )

    try:
        _download_with_retries(
            safe_url,
            destination,
            timeout=max(DEFAULT_DOWNLOAD_TIMEOUT, 1.0),
            chunk_size=max(DEFAULT_CHUNK_SIZE, 65536),
            retries=3,
        )
    except Exception as exc:  # noqa: BLE001
        LOGGER.error(
            "file_get download failed",
            extra={
                "file_id": graph_file_id,
                "account_id": account,
                "destination": str(destination),
            },
        )
        if destination.exists():
            destination.unlink(missing_ok=True)
        if isinstance(exc, ValidationError):
            raise
        raise RuntimeError(f"Failed to download file: {exc}") from exc

    actual_size = destination.stat().st_size
    return {
        "path": str(destination),
        "name": metadata.get("name", "unknown"),
        "size_mb": round(actual_size / (1024 * 1024), 2),
        "mime_type": metadata.get("file", {}).get("mimeType") if metadata else None,
    }


@mcp.tool(
    name="file_get",
    annotations={
        "title": "Get File",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={"category": "file", "safety_level": "moderate"},
)
def file_get(file_id: str, account_id: str, download_path: str) -> dict[str, Any]:
    """✏️ Download a OneDrive file to a local path (requires user confirmation recommended)

    The download URL is supplied by Microsoft Graph (never user input) and is
    validated against an allow-list of Microsoft domains before use. The file
    is streamed to disk in configurable chunks with retry behaviour to protect
    against transient failures. Download size and timeouts respect the
    environment variables `MCP_FILE_DOWNLOAD_MAX_MB` and
    `MCP_FILE_DOWNLOAD_TIMEOUT`.

    Args:
        file_id: The Microsoft Graph file identifier to download.
        account_id: Microsoft account identifier associated with the file.
        download_path: Absolute path where the file will be stored locally.
            Must reside within an allowed root directory.

    Returns:
        Dictionary containing download metadata (name, size_mb, mime_type).

    Raises:
        ValidationError: If input parameters are invalid.
        RuntimeError: If all download attempts fail.
    """

    return _file_get_impl(file_id, account_id, download_path)


# file_create
@mcp.tool(
    name="file_create",
    annotations={
        "title": "Create File",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": False,
        "openWorldHint": True,
    },
    meta={"category": "file", "safety_level": "moderate"},
)
def file_create(
    onedrive_path: str, local_file_path: str, account_id: str
) -> dict[str, Any]:
    """✏️ Upload a local file to OneDrive (requires user confirmation recommended)

    Args:
        onedrive_path: Destination path within OneDrive (must start with '/').
        local_file_path: Absolute path to the local file to upload. Paths are
            validated via `ensure_safe_path` to prevent traversal and
            restrict uploads to trusted directories.
        account_id: Microsoft account identifier.

    Returns:
        Metadata for the created OneDrive file.
    """
    account = validate_account_id(account_id)
    target = validate_onedrive_path(onedrive_path)
    source_path = ensure_safe_path(
        local_file_path, must_exist=True, allow_overwrite=True
    )

    data = source_path.read_bytes()
    result = graph.upload_large_file(f"/me/drive/root:{target}:", data, account)
    if not result:
        raise RuntimeError(f"Failed to create file at path: {target}")
    return result


# file_update
@mcp.tool(
    name="file_update",
    annotations={
        "title": "Update File",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={"category": "file", "safety_level": "moderate"},
)
def file_update(file_id: str, local_file_path: str, account_id: str) -> dict[str, Any]:
    """✏️ Replace OneDrive file content with local file (requires user confirmation recommended)

    Args:
        file_id: Target OneDrive file identifier to replace.
        local_file_path: Absolute path to the replacement file. Validated via
            `ensure_safe_path` to block traversal and enforce workspace roots.
        account_id: Microsoft account identifier.

    Returns:
        Updated file metadata returned by Microsoft Graph.
    """
    account = validate_account_id(account_id)
    graph_file_id = validate_microsoft_graph_id(file_id, "file_id")
    source_path = ensure_safe_path(
        local_file_path, must_exist=True, allow_overwrite=True
    )
    data = source_path.read_bytes()
    result = graph.upload_large_file(
        f"/me/drive/items/{graph_file_id}",
        data,
        account,
    )
    if not result:
        raise RuntimeError(f"Failed to update file with ID: {graph_file_id}")
    return result


# file_delete
@mcp.tool(
    name="file_delete",
    annotations={
        "title": "Delete File",
        "readOnlyHint": False,
        "destructiveHint": True,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={
        "category": "file",
        "safety_level": "critical",
        "requires_confirmation": True,
    },
)
def file_delete(file_id: str, account_id: str, confirm: bool = False) -> dict[str, str]:
    """🔴 Delete a OneDrive file or folder permanently (always require user confirmation)

    WARNING: This action permanently deletes the file or folder and cannot be undone.
    """

    require_confirm(confirm, "delete OneDrive item")
    account = validate_account_id(account_id)
    graph_file_id = validate_microsoft_graph_id(file_id, "file_id")
    graph.request("DELETE", f"/me/drive/items/{graph_file_id}", account)
    return {"status": "deleted"}


def _list_folders_impl(
    account_id: str,
    path: str | None = "/",
    folder_id: str | None = None,
    limit: int | None = None,
) -> list[dict[str, Any]]:
    """Internal implementation for listing OneDrive folders"""
    if folder_id:
        endpoint = f"/me/drive/items/{folder_id}/children"
    elif path in (None, "/"):
        endpoint = "/me/drive/root/children"
    else:
        endpoint = f"/me/drive/root:/{path}:/children"

    page_size = limit if limit is not None else 500
    params = {
        "$top": page_size,
        "$select": "id,name,folder,parentReference,size,lastModifiedDateTime",
    }

    items = list(
        graph.request_paginated(endpoint, account_id, params=params, limit=limit)
    )

    # Filter to only return folders
    folders = []
    for item in items:
        if "folder" in item:
            folders.append(
                {
                    "id": item["id"],
                    "name": item["name"],
                    "childCount": item.get("folder", {}).get("childCount", 0),
                    "path": item.get("parentReference", {}).get("path", ""),
                    "parentId": item.get("parentReference", {}).get("id"),
                    "modified": item.get("lastModifiedDateTime"),
                }
            )

    return folders



from __future__ import annotations

from datetime import timezone
from typing import Any
from zoneinfo import ZoneInfo

from .. import graph
from ..mcp_instance import mcp
from ..validators import (
    ValidationError,
    format_validation_error,
    normalize_recipients,
    require_confirm,
    validate_choices,
    validate_datetime_window,
    validate_iso_datetime,
    validate_json_payload,
    validate_timezone,
)

MAX_CALENDAR_ATTENDEES = 500
CALENDAR_RESPONSE_ALIASES = {
    "accept": "accept",
    "decline": "decline",
    "tentativelyAccept": "tentativelyAccept",
    "tentative": "tentativelyAccept",
}
ALLOWED_CALENDAR_UPDATE_KEYS = (
    "subject",
    "start",
    "end",
    "timezone",
    "location",
    "body",
    "attendees",
)


# calendar_get_event
@mcp.tool(
    name="calendar_get_event",
    annotations={
        "title": "Get Calendar Event",
        "readOnlyHint": True,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={"category": "calendar", "safety_level": "safe"},
)
def calendar_get_event(event_id: str, account_id: str) -> dict[str, Any]:
    """📖 Get full event details (read-only, safe for unsupervised use)

    Returns complete event information including recurrence patterns and online meeting details.

    Args:
        event_id: The event ID
        account_id: Microsoft account ID

    Returns:
        Complete event object with all metadata
    """
    result = graph.request("GET", f"/me/events/{event_id}", account_id)
    if not result:
        raise ValueError(f"Event with ID {event_id} not found")
    return result


# calendar_create_event
@mcp.tool(
    name="calendar_create_event",
    annotations={
        "title": "Create Calendar Event",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": False,
        "openWorldHint": True,
    },
    meta={"category": "calendar", "safety_level": "moderate"},
)
def calendar_create_event(
    account_id: str,
    subject: str,
    start: str,
    end: str,
    location: str | None = None,
    body: str | None = None,
    attendees: str | list[str] | None = None,
    timezone: str = "UTC",
) -> dict[str, Any]:
    """✏️ Create a calendar event (requires user confirmation recommended)

    Creates a new calendar event with optional attendees and location.
    Attendees will receive meeting invitations. Addresses are validated,
    deduplicated, and limited to 500 unique recipients.

    Args:
        account_id: Microsoft account ID
        subject: Event title
        start: Start time in ISO format (e.g., "2024-01-15T10:00:00")
        end: End time in ISO format
        location: Location name (optional)
        body: Event description (optional)
        attendees: Email address(es) of attendees (optional)
        timezone: Timezone for the event (default: "UTC")

    Returns:
        Created event object with ID

    Raises:
        ValidationError: If datetime values, timezone, or attendee
            addresses are invalid.
    """
    start_dt, end_dt = validate_datetime_window(start, end)
    timezone_normalized = validate_timezone(timezone)
    tzinfo = ZoneInfo(timezone_normalized)
    start_local = start_dt.astimezone(tzinfo)
    end_local = end_dt.astimezone(tzinfo)

    attendees_deduped: list[str] = []
    if attendees:
        attendee_candidates = normalize_recipients(attendees, "attendees")
        seen: set[str] = set()
        for address in attendee_candidates:
            key = address.casefold()
            if key in seen:
                continue
            seen.add(key)
            attendees_deduped.append(address)
        if len(attendees_deduped) > MAX_CALENDAR_ATTENDEES:
            reason = (
                f"must not exceed {MAX_CALENDAR_ATTENDEES} unique attendees per event"
            )
            raise ValidationError(
                format_validation_error(
                    "attendees",
                    len(attendees_deduped),
                    reason,
                    f"≤ {MAX_CALENDAR_ATTENDEES}",
                )
            )

    event = {
        "subject": subject,
        "start": {
            "dateTime": start_local.isoformat(),
            "timeZone": timezone_normalized,
        },
        "end": {
            "dateTime": end_local.isoformat(),
            "timeZone": timezone_normalized,
        },
    }

    if location:
        event["location"] = {"displayName": location}

    if body:
        event["body"] = {"contentType": "Text", "content": body}

    if attendees_deduped:
        event["attendees"] = [
            {"emailAddress": {"address": address}, "type": "required"}
            for address in attendees_deduped
        ]

    result = graph.request("POST", "/me/events", account_id, json=event)
    if not result:
        raise ValueError("Failed to create event")
    return result


# calendar_update_event
@mcp.tool(
    name="calendar_update_event",
    annotations={
        "title": "Update Calendar Event",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={"category": "calendar", "safety_level": "moderate"},
)
def calendar_update_event(
    event_id: str, updates: dict[str, Any], account_id: str
) -> dict[str, Any]:
    """✏️ Update event properties (requires user confirmation recommended)

    Modifies event details like time, location, or attendees.
    Attendees will receive update notifications.

    Allowed update keys: subject, start, end, timezone, location, body, attendees.

    Args:
        event_id: The event ID to update
        updates: Dictionary with fields to update (subject, start, end, location, body)
        account_id: Microsoft account ID

    Returns:
        Updated event object
    """
    payload = validate_json_payload(
        updates,
        allowed_keys=ALLOWED_CALENDAR_UPDATE_KEYS,
        param_name="updates",
    )
    if not payload:
        raise ValidationError(
            format_validation_error(
                "updates",
                payload,
                "must include at least one field",
                f"Subset of {sorted(ALLOWED_CALENDAR_UPDATE_KEYS)}",
            )
        )

    timezone_override: str | None = None
    if "timezone" in payload:
        timezone_override = validate_timezone(payload["timezone"], "updates.timezone")
        if "start" not in payload and "end" not in payload:
            raise ValidationError(
                format_validation_error(
                    "updates.timezone",
                    payload["timezone"],
                    "can only be provided when updating start or end",
                    "Include start/end alongside timezone",
                )
            )

    start_value: str | None = None
    end_value: str | None = None

    if "start" in payload:
        start_raw = payload["start"]
        if not isinstance(start_raw, str):
            raise ValidationError(
                format_validation_error(
                    "updates.start",
                    start_raw,
                    "must be an ISO-8601 datetime string",
                    "YYYY-MM-DDTHH:MM:SS+TZ",
                )
            )
        validate_iso_datetime(start_raw, "updates.start")
        start_value = start_raw

    if "end" in payload:
        end_raw = payload["end"]
        if not isinstance(end_raw, str):
            raise ValidationError(
                format_validation_error(
                    "updates.end",
                    end_raw,
                    "must be an ISO-8601 datetime string",
                    "YYYY-MM-DDTHH:MM:SS+TZ",
                )
            )
        validate_iso_datetime(end_raw, "updates.end")
        end_value = end_raw

    if start_value and end_value:
        validate_datetime_window(start_value, end_value)

    formatted_updates: dict[str, Any] = {}

    if "subject" in payload:
        subject_value = payload["subject"]
        if not isinstance(subject_value, str):
            raise ValidationError(
                format_validation_error(
                    "updates.subject",
                    subject_value,
                    "must be a string",
                    "Subject string",
                )
            )
        formatted_updates["subject"] = subject_value

    timezone_for_dates = timezone_override or "UTC"

    if start_value is not None:
        formatted_updates["start"] = {
            "dateTime": start_value,
            "timeZone": timezone_for_dates,
        }

    if end_value is not None:
        formatted_updates["end"] = {
            "dateTime": end_value,
            "timeZone": timezone_for_dates,
        }

    if "location" in payload:
        location_value = payload["location"]
        if not isinstance(location_value, str):
            raise ValidationError(
                format_validation_error(
                    "updates.location",
                    location_value,
                    "must be a string",
                    "Location name string",
                )
            )
        formatted_updates["location"] = {"displayName": location_value.strip()}

    if "body" in payload:
        body_value = payload["body"]
        if not isinstance(body_value, str):
            raise ValidationError(
                format_validation_error(
                    "updates.body",
                    body_value,
                    "must be a string",
                    "Event description string",
                )
            )
        formatted_updates["body"] = {"contentType": "Text", "content": body_value}

    if "attendees" in payload:
        attendees_value = payload["attendees"]
        if isinstance(attendees_value, list) and not attendees_value:
            formatted_updates["attendees"] = []
        else:
            attendee_candidates = normalize_recipients(
                attendees_value,
                "updates.attendees",
            )
            deduped: list[str] = []
            seen: set[str] = set()
            for address in attendee_candidates:
                key = address.casefold()
                if key in seen:
                    continue
                seen.add(key)
                deduped.append(address)
            if len(deduped) > MAX_CALENDAR_ATTENDEES:
                raise ValidationError(
                    format_validation_error(
                        "updates.attendees",
                        len(deduped),
                        f"must not exceed {MAX_CALENDAR_ATTENDEES} attendees",
                        f"≤ {MAX_CALENDAR_ATTENDEES}",
                    )
                )
            formatted_updates["attendees"] = [
                {"emailAddress": {"address": address}, "type": "required"}
                for address in deduped
            ]

    result = graph.request(
        "PATCH", f"/me/events/{event_id}", account_id, json=formatted_updates
    )
    return result or {"status": "updated"}


# calendar_delete_event
@mcp.tool(
    name="calendar_delete_event",
    annotations={
        "title": "Delete Calendar Event",
        "readOnlyHint": False,
        "destructiveHint": True,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={
        "category": "calendar",
        "safety_level": "critical",
        "requires_confirmation": True,
    },
)
def calendar_delete_event(
    account_id: str,
    event_id: str,
    send_cancellation: bool = True,
    confirm: bool = False,
) -> dict[str, str]:
    """🔴 Delete a calendar event (always require user confirmation)

    WARNING: This action permanently deletes the event and cannot be undone.
    If this is a meeting, attendees will receive cancellation notices.

    Args:
        account_id: Microsoft account ID
        event_id: The event to delete
        send_cancellation: Whether to notify attendees (default: True)
        confirm: Must be True to confirm deletion (prevents accidents)

    Returns:
        Status confirmation
    """
    require_confirm(confirm, "delete calendar event")
    if send_cancellation:
        graph.request("POST", f"/me/events/{event_id}/cancel", account_id, json={})
    else:
        graph.request("DELETE", f"/me/events/{event_id}", account_id)
    return {"status": "deleted"}


# calendar_respond_event
@mcp.tool(
    name="calendar_respond_event",
    annotations={
        "title": "Respond to Calendar Event",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={"category": "calendar", "safety_level": "moderate"},
)
def calendar_respond_event(
    account_id: str,
    event_id: str,
    response: str = "accept",
    message: str | None = None,
) -> dict[str, str]:
    """⚠️ Respond to a calendar event invitation (requires user confirmation recommended)

    IMPORTANT: This sends a response to the event organizer.

    Valid responses:
        - "accept" - Accept the invitation
        - "decline" - Decline the invitation
        - "tentativelyAccept" - Mark as tentative
      (Input is case-insensitive; "tentative" is accepted as an alias.)

    Args:
        account_id: Microsoft account ID
        event_id: The event ID to respond to
        response: Response type (default: "accept")
        message: Optional message to the organizer

    Returns:
        Status confirmation

    Raises:
        ValidationError: If the response value or message payload is invalid.
    """
    canonical_key = validate_choices(
        response,
        CALENDAR_RESPONSE_ALIASES.keys(),
        "response",
    )
    resolved_response = CALENDAR_RESPONSE_ALIASES[canonical_key]

    payload: dict[str, Any] = {"sendResponse": True}

    if message is not None:
        if not isinstance(message, str):
            reason = "must be a string"
            raise ValidationError(
                format_validation_error(
                    "message",
                    message,
                    reason,
                    "Non-empty response message or omit parameter",
                )
            )
        message_trimmed = message.strip()
        if not message_trimmed:
            reason = "cannot be empty when provided"
            raise ValidationError(
                format_validation_error(
                    "message",
                    message,
                    reason,
                    "Non-empty response message or omit parameter",
                )
            )
        payload["comment"] = message_trimmed

    graph.request(
        "POST",
        f"/me/events/{event_id}/{resolved_response}",
        account_id,
        json=payload,
    )
    return {"status": resolved_response}


# calendar_check_availability
@mcp.tool(
    name="calendar_check_availability",
    annotations={
        "title": "Check Calendar Availability",
        "readOnlyHint": True,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={"category": "calendar", "safety_level": "safe"},
)
def calendar_check_availability(
    account_id: str,
    start: str,
    end: str,
    attendees: str | list[str] | None = None,
) -> dict[str, Any]:
    """📖 Check calendar availability for scheduling (read-only, safe for unsupervised use)

    Returns free/busy information for the user and optional attendees.
    Useful for finding meeting times.

    Args:
        account_id: Microsoft account ID
        start: Start time in ISO format
        end: End time in ISO format
        attendees: Optional email address(es) to check availability

    Returns:
        Schedule information with free/busy slots

    Raises:
        ValidationError: If start/end datetimes or attendee addresses
            are invalid.
        ValueError: If the current account email address is unavailable.
    """
    start_dt, end_dt = validate_datetime_window(start, end)

    attendee_addresses: list[str] = []
    if attendees:
        attendee_candidates = normalize_recipients(attendees, "attendees")
        seen_attendees: set[str] = set()
        for address in attendee_candidates:
            key = address.casefold()
            if key in seen_attendees:
                continue
            seen_attendees.add(key)
            attendee_addresses.append(address)
        if len(attendee_addresses) > MAX_CALENDAR_ATTENDEES:
            reason = (
                f"must not exceed {MAX_CALENDAR_ATTENDEES} unique attendees per request"
            )
            raise ValidationError(
                format_validation_error(
                    "attendees",
                    len(attendee_addresses),
                    reason,
                    f"≤ {MAX_CALENDAR_ATTENDEES}",
                )
            )

    me_info = graph.request("GET", "/me", account_id)
    if not me_info or "mail" not in me_info:
        raise ValueError("Failed to get user email address")
    schedules = [me_info["mail"]]
    current_keys = {me_info["mail"].casefold()}
    for address in attendee_addresses:
        key = address.casefold()
        if key in current_keys:
            continue
        current_keys.add(key)
        schedules.append(address)

    payload = {
        "schedules": schedules,
        "startTime": {
            "dateTime": start_dt.astimezone(timezone.utc).isoformat(),
            "timeZone": "UTC",
        },
        "endTime": {
            "dateTime": end_dt.astimezone(timezone.utc).isoformat(),
            "timeZone": "UTC",
        },
        "availabilityViewInterval": 30,
    }

    result = graph.request("POST", "/me/calendar/getSchedule", account_id, json=payload)
    if not result:
        raise ValueError("Failed to check availability")
    return result



from __future__ import annotations

import subprocess
import time
from collections import deque
from pathlib import Path
from typing import Any, Callable

import pytest

from src.microsoft_mcp import graph


@pytest.fixture
def mock_graph_request(
    monkeypatch: pytest.MonkeyPatch,
) -> Callable[[str, str, Any], None]:
    """Patch graph.request and allow registering responses."""

    responses: dict[tuple[str, str], Any] = {}

    def register(method: str, path: str, response: Any) -> None:
        responses[(method, path)] = response

    def fake_request(
        method: str,
        path: str,
        account_id: str | None = None,
        **kwargs: Any,
    ) -> Any:
        key = (method, path)
        if key not in responses:
            raise AssertionError(f"Unexpected Graph request: {method} {path}")
        value = responses[key]
        return value() if callable(value) else value

    monkeypatch.setattr(graph, "request", fake_request)
    return register


@pytest.fixture
def mock_graph_paginated(
    monkeypatch: pytest.MonkeyPatch,
) -> Callable[[str, list[dict[str, Any]]], None]:
    """Patch graph.request_paginated with configurable payloads."""

    payloads: dict[str, deque[list[dict[str, Any]]]] = {}

    def register(path: str, pages: list[dict[str, Any]]) -> None:
        payloads.setdefault(path, deque()).append(pages)

    def fake_paginated(
        path: str,
        account_id: str | None = None,
        params: dict[str, Any] | None = None,
        limit: int | None = None,
    ):
        queue = payloads.get(path)
        if not queue:
            return iter(())
        page = queue.popleft()
        return iter(page)

    monkeypatch.setattr(graph, "request_paginated", fake_paginated)
    return register


@pytest.fixture
def mock_account_id() -> str:
    """Provide a reusable fake account identifier."""

    return "test-account"


@pytest.fixture
def mock_email_data() -> Callable[..., dict[str, Any]]:
    """Factory for building fake email payloads."""

    def _builder(**overrides: Any) -> dict[str, Any]:
        payload = {
            "id": "mock-email-id",
            "subject": "Test Subject",
            "body": {"contentType": "Text", "content": "Hello"},
            "hasAttachments": False,
        }
        payload.update(overrides)
        return payload

    return _builder


@pytest.fixture
def mock_file_metadata() -> Callable[..., dict[str, Any]]:
    """Factory for building fake OneDrive file metadata payloads."""

    def _builder(**overrides: Any) -> dict[str, Any]:
        payload: dict[str, Any] = {
            "id": "01ABCDEFZLMNO!123",
            "name": "example.txt",
            "size": 2048,
            "file": {"mimeType": "text/plain"},
            "@microsoft.graph.downloadUrl": (
                "https://graph.microsoft.com/v1.0/me/drive/items/01ABCDEFZLMNO!123/content"
            ),
        }
        payload.update(overrides)
        return payload

    return _builder


@pytest.fixture
def mock_calendar_event() -> Callable[..., dict[str, Any]]:
    """Factory for creating calendar event payloads."""

    def _builder(**overrides: Any) -> dict[str, Any]:
        payload: dict[str, Any] = {
            "id": "event-123",
            "subject": "Team Sync",
            "start": {"dateTime": "2024-05-01T10:00:00+00:00", "timeZone": "UTC"},
            "end": {"dateTime": "2024-05-01T11:00:00+00:00", "timeZone": "UTC"},
            "attendees": [
                {
                    "emailAddress": {"address": "owner@example.com", "name": "Owner"},
                    "type": "required",
                }
            ],
        }
        payload.update(overrides)
        return payload

    return _builder


@pytest.fixture
def mock_contact_data() -> Callable[..., dict[str, Any]]:
    """Factory for creating contact payloads."""

    def _builder(**overrides: Any) -> dict[str, Any]:
        payload: dict[str, Any] = {
            "id": "contact-123",
            "displayName": "Ada Lovelace",
            "givenName": "Ada",
            "surname": "Lovelace",
            "emailAddresses": [
                {"address": "ada.lovelace@example.com", "name": "Ada Lovelace"}
            ],
        }
        payload.update(overrides)
        return payload

    return _builder


@pytest.fixture
def temp_safe_dir(tmp_path: Path) -> Path:
    """Provide an isolated directory for filesystem interactions."""

    target = tmp_path / "safe"
    target.mkdir()
    return target


@pytest.fixture(autouse=True, scope="session")
def ensure_port_8000_free():
    """Kill any process on port 8000 before and after test session.

    This prevents orphaned HTTP server processes from blocking the port
    and causing test failures. Runs automatically for every test session.
    """

    def cleanup():
        """Find and kill all processes using port 8000 using netstat approach."""
        try:
            # Use the proven netstat approach to kill processes on port 8000
            # Loop while port 8000 is in use (max 3 iterations to avoid infinite loop)
            max_attempts = 3
            attempt = 0

            while attempt < max_attempts:
                # Check if port 8000 is in use
                check_result = subprocess.run(
                    ["sudo", "netstat", "-tunlp"],
                    capture_output=True,
                    text=True,
                    timeout=5,
                    check=False,
                )

                if "8000" not in check_result.stdout:
                    # Port is free
                    break

                # Extract PID and kill it
                # Parse: "tcp 0 0 :::8000 :::* LISTEN 12345/python"
                # Extract: "12345/python" then remove last 8 chars to get PID
                for line in check_result.stdout.split("\n"):
                    if "8000" in line:
                        # Use awk to get column 7, sed to remove last 8 chars
                        pid_result = subprocess.run(
                            ["bash", "-c", f"echo '{line}' | awk '{{print $7}}' | sed 's/.{{8}}$//'"],
                            capture_output=True,
                            text=True,
                            timeout=2,
                            check=False,
                        )
                        pid = pid_result.stdout.strip()

                        if pid:
                            # Kill the process
                            subprocess.run(
                                ["sudo", "kill", "-9", pid],
                                stderr=subprocess.DEVNULL,
                                timeout=2,
                                check=False,
                            )

                # Wait before next iteration
                time.sleep(1)
                attempt += 1

        except subprocess.TimeoutExpired:
            print("Warning: Port 8000 cleanup timed out")
        except Exception as e:
            # Don't fail tests if cleanup fails, just warn
            print(f"Warning: Port 8000 cleanup encountered error: {e}")

    # Clean before test session starts
    cleanup()

    yield

    # Clean after test session ends
    cleanup()



from __future__ import annotations

import pytest

from src.microsoft_mcp.tools import email_folders as email_folder_tools
from src.microsoft_mcp.tools import file as file_tools
from src.microsoft_mcp.tools import folder as folder_tools
from src.microsoft_mcp.validators import ValidationError


def test_folder_list_rejects_invalid_limit(mock_account_id: str) -> None:
    with pytest.raises(ValidationError):
        folder_tools.folder_list.fn(account_id=mock_account_id, limit=0)


def test_folder_get_tree_rejects_invalid_depth(mock_account_id: str) -> None:
    with pytest.raises(ValidationError):
        folder_tools.folder_get_tree.fn(
            account_id=mock_account_id,
            max_depth=0,
        )


def test_emailfolders_list_rejects_invalid_limit(mock_account_id: str) -> None:
    with pytest.raises(ValidationError):
        email_folder_tools.emailfolders_list.fn(account_id=mock_account_id, limit=0)


def test_emailfolders_get_tree_rejects_invalid_depth(mock_account_id: str) -> None:
    with pytest.raises(ValidationError):
        email_folder_tools.emailfolders_get_tree.fn(
            account_id=mock_account_id,
            max_depth=0,
        )


def test_file_list_rejects_relative_path(mock_account_id: str) -> None:
    with pytest.raises(ValidationError):
        file_tools.file_list.fn(
            account_id=mock_account_id,
            path="Documents",
        )


def test_folder_list_rejects_relative_path(mock_account_id: str) -> None:
    with pytest.raises(ValidationError):
        folder_tools.folder_list.fn(
            account_id=mock_account_id,
            path="Documents/Work",
        )


def test_folder_get_rejects_relative_path(mock_account_id: str) -> None:
    with pytest.raises(ValidationError):
        folder_tools.folder_get.fn(
            account_id=mock_account_id,
            path="Projects/2024",
        )



from ..mcp_instance import mcp
from importlib.metadata import version, PackageNotFoundError


# server_get_version
@mcp.tool(
    name="server_get_version",
    annotations={
        "title": "Get Server Version",
        "readOnlyHint": True,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={"category": "server", "safety_level": "safe"},
)
def server_get_version() -> dict[str, str]:
    """📖 Get the version of the m365-mcp server (read-only, safe for unsupervised use)

    Returns the current version of the m365-mcp server that is running.
    Useful for diagnostics, troubleshooting, and ensuring compatibility.

    Returns:
        Dictionary containing:
        - version: The semantic version string (e.g., "0.1.3")
        - package: The package name ("m365-mcp")

    Example:
        >>> server_get_version()
        {"version": "0.1.3", "package": "m365-mcp"}
    """
    try:
        pkg_version = version("m365-mcp")
    except PackageNotFoundError:
        # Fallback for development environments where package isn't installed
        pkg_version = "dev"

    return {
        "package": "m365-mcp",
        "version": pkg_version,
    }



"""
Microsoft MCP Tools Registry

This module serves as the central registry for all Microsoft MCP tools.
It imports the FastMCP instance from mcp_instance and imports all tool modules
to ensure all tools are registered with the same instance.

The actual tool implementations are in the tools/ subdirectory, organized by category:
- tools/account.py - Account management (3 tools)
- tools/calendar.py - Calendar operations (6 tools)
- tools/contact.py - Contact management (5 tools)
- tools/email.py - Email operations (9 tools)
- tools/email_folders.py - Email folder navigation (3 tools)
- tools/email_rules.py - Email rule management (9 tools)
- tools/file.py - OneDrive file operations (5 tools)
- tools/folder.py - OneDrive folder navigation (3 tools)
- tools/search.py - Search operations (5 tools)

Total: 50 tools across 9 modules
"""

# Import the mcp instance
from .mcp_instance import mcp

# Import all tool modules - this triggers tool registration via @mcp.tool decorators
# The tools themselves are defined in the tools/ subdirectory

# Export the mcp instance for use by server.py
__all__ = ["mcp"]



from typing import Any
from ..mcp_instance import mcp
from .. import graph
from ..validators import (
    ValidationError,
    format_validation_error,
    require_confirm,
    validate_choices,
    validate_email_format,
    validate_json_payload,
)

RULE_FLAG_STATUS_CHOICES = {"notFlagged", "flagged", "complete"}
RULE_IMPORTANCE_CHOICES = {"low", "normal", "high"}
RULE_SENSITIVITY_CHOICES = {"normal", "personal", "private", "confidential"}
RULE_MEETING_TYPES = {
    "meetingRequest",
    "meetingCancelled",
    "meetingAccepted",
    "meetingTentativelyAccepted",
    "meetingDeclined",
}

RULE_PREDICATE_STRING_LIST_KEYS = {
    "categories",
    "subjectContains",
    "bodyContains",
    "bodyOrSubjectContains",
    "senderContains",
    "recipientContains",
    "senderAddressContains",
    "recipientAddressContains",
    "headers",
    "headerContains",
    "searchTerms",
}
RULE_PREDICATE_RECIPIENT_KEYS = {
    "fromAddresses",
    "sentToAddresses",
}
RULE_PREDICATE_BOOL_KEYS = {
    "hasAttachments",
    "stopProcessingRules",
}
RULE_PREDICATE_CHOICE_MAP = {
    "importance": RULE_IMPORTANCE_CHOICES,
    "sensitivity": RULE_SENSITIVITY_CHOICES,
}
ALLOWED_RULE_PREDICATE_KEYS = tuple(
    RULE_PREDICATE_STRING_LIST_KEYS
    | RULE_PREDICATE_RECIPIENT_KEYS
    | RULE_PREDICATE_BOOL_KEYS
    | set(RULE_PREDICATE_CHOICE_MAP.keys())
    | {
        "withinSizeRange",
        "messageActionFlag",
        "meetingMessageType",
    }
)

RULE_ACTION_STRING_LIST_KEYS = {"assignCategories"}
RULE_ACTION_STRING_KEYS = {"copyToFolder", "moveToFolder"}
RULE_ACTION_RECIPIENT_KEYS = {
    "forwardTo",
    "forwardAsAttachmentTo",
    "redirectTo",
}
RULE_ACTION_BOOL_KEYS = {
    "markAsRead",
    "stopProcessingRules",
    "delete",
    "permanentDelete",
}
RULE_ACTION_CHOICE_MAP = {"markImportance": RULE_IMPORTANCE_CHOICES}
ALLOWED_RULE_ACTION_KEYS = tuple(
    RULE_ACTION_STRING_LIST_KEYS
    | RULE_ACTION_STRING_KEYS
    | RULE_ACTION_RECIPIENT_KEYS
    | RULE_ACTION_BOOL_KEYS
    | set(RULE_ACTION_CHOICE_MAP.keys())
)


def _validate_bool(value: Any, param_name: str) -> bool:
    if not isinstance(value, bool):
        raise ValidationError(
            format_validation_error(
                param_name,
                value,
                "must be a boolean",
                "True or False",
            )
        )
    return value


def _validate_string_list(values: Any, param_name: str) -> list[str]:
    if not isinstance(values, list):
        raise ValidationError(
            format_validation_error(
                param_name,
                values,
                "must be a list of strings",
                "List of non-empty strings",
            )
        )
    normalised: list[str] = []
    for index, item in enumerate(values):
        if not isinstance(item, str):
            raise ValidationError(
                format_validation_error(
                    f"{param_name}[{index}]",
                    item,
                    "must be a string",
                    "Text value",
                )
            )
        trimmed = item.strip()
        if not trimmed:
            raise ValidationError(
                format_validation_error(
                    f"{param_name}[{index}]",
                    item,
                    "cannot be empty",
                    "Non-empty string",
                )
            )
        normalised.append(trimmed)
    return normalised


def _normalise_rule_recipients(
    recipients: Any,
    param_name: str,
) -> list[dict[str, dict[str, str]]]:
    if not isinstance(recipients, list):
        raise ValidationError(
            format_validation_error(
                param_name,
                recipients,
                "must be a list of recipient objects",
                "List of email addresses",
            )
        )
    result: list[dict[str, dict[str, str]]] = []
    for index, entry in enumerate(recipients):
        if isinstance(entry, str):
            email = validate_email_format(
                entry,
                f"{param_name}[{index}]",
            )
            result.append({"emailAddress": {"address": email}})
            continue
        if not isinstance(entry, dict):
            raise ValidationError(
                format_validation_error(
                    f"{param_name}[{index}]",
                    entry,
                    "must be a recipient dictionary",
                    "{'emailAddress': {'address': str}}",
                )
            )
        if "emailAddress" in entry and isinstance(entry["emailAddress"], dict):
            email_address = entry["emailAddress"].get("address")
            if not isinstance(email_address, str):
                raise ValidationError(
                    format_validation_error(
                        f"{param_name}[{index}].emailAddress.address",
                        email_address,
                        "must be a string",
                        "Email address",
                    )
                )
            email = validate_email_format(
                email_address,
                f"{param_name}[{index}].emailAddress.address",
            )
            result.append({"emailAddress": {"address": email}})
            continue
        if "address" in entry:
            email = validate_email_format(
                entry["address"],
                f"{param_name}[{index}].address",
            )
            result.append({"emailAddress": {"address": email}})
            continue
        raise ValidationError(
            format_validation_error(
                f"{param_name}[{index}]",
                entry,
                "must include an email address",
                "{'emailAddress': {'address': str}}",
            )
        )
    return result


def _validate_size_range(value: Any, param_name: str) -> dict[str, int]:
    payload = validate_json_payload(
        value,
        allowed_keys=("minimumSize", "maximumSize"),
        param_name=param_name,
    )
    validated: dict[str, int] = {}
    for key in ("minimumSize", "maximumSize"):
        if key in payload:
            amount = payload[key]
            if not isinstance(amount, int):
                raise ValidationError(
                    format_validation_error(
                        f"{param_name}.{key}",
                        amount,
                        "must be an integer number of bytes",
                        "Integer value ≥ 0",
                    )
                )
            if amount < 0:
                raise ValidationError(
                    format_validation_error(
                        f"{param_name}.{key}",
                        amount,
                        "cannot be negative",
                        "Value ≥ 0",
                    )
                )
            validated[key] = amount
    if (
        "minimumSize" in validated
        and "maximumSize" in validated
        and validated["minimumSize"] > validated["maximumSize"]
    ):
        raise ValidationError(
            format_validation_error(
                param_name,
                value,
                "minimumSize cannot exceed maximumSize",
                "minimumSize ≤ maximumSize",
            )
        )
    return validated


def _validate_rule_predicates(
    predicates: Any,
    param_name: str,
) -> dict[str, Any]:
    payload = validate_json_payload(
        predicates,
        allowed_keys=ALLOWED_RULE_PREDICATE_KEYS,
        param_name=param_name,
    )
    validated: dict[str, Any] = {}
    for key, value in payload.items():
        if key in RULE_PREDICATE_STRING_LIST_KEYS:
            validated[key] = _validate_string_list(value, f"{param_name}.{key}")
        elif key in RULE_PREDICATE_RECIPIENT_KEYS:
            validated[key] = _normalise_rule_recipients(value, f"{param_name}.{key}")
        elif key in RULE_PREDICATE_BOOL_KEYS:
            validated[key] = _validate_bool(value, f"{param_name}.{key}")
        elif key in RULE_PREDICATE_CHOICE_MAP:
            validated[key] = validate_choices(
                value,
                RULE_PREDICATE_CHOICE_MAP[key],
                f"{param_name}.{key}",
            )
        elif key == "withinSizeRange":
            validated[key] = _validate_size_range(value, f"{param_name}.{key}")
        elif key == "messageActionFlag":
            validated[key] = validate_choices(
                value,
                RULE_FLAG_STATUS_CHOICES,
                f"{param_name}.messageActionFlag",
            )
        elif key == "meetingMessageType":
            validated[key] = validate_choices(
                value,
                RULE_MEETING_TYPES,
                f"{param_name}.meetingMessageType",
            )
    return validated


def _validate_rule_actions(
    actions: Any,
    param_name: str,
) -> dict[str, Any]:
    payload = validate_json_payload(
        actions,
        allowed_keys=ALLOWED_RULE_ACTION_KEYS,
        param_name=param_name,
    )
    validated: dict[str, Any] = {}
    for key, value in payload.items():
        if key in RULE_ACTION_STRING_LIST_KEYS:
            validated[key] = _validate_string_list(value, f"{param_name}.{key}")
        elif key in RULE_ACTION_STRING_KEYS:
            if not isinstance(value, str):
                raise ValidationError(
                    format_validation_error(
                        f"{param_name}.{key}",
                        value,
                        "must be a string",
                        "Identifier or folder ID string",
                    )
                )
            trimmed = value.strip()
            if not trimmed:
                raise ValidationError(
                    format_validation_error(
                        f"{param_name}.{key}",
                        value,
                        "cannot be empty",
                        "Non-empty string",
                    )
                )
            validated[key] = trimmed
        elif key in RULE_ACTION_RECIPIENT_KEYS:
            validated[key] = _normalise_rule_recipients(value, f"{param_name}.{key}")
        elif key in RULE_ACTION_BOOL_KEYS:
            validated[key] = _validate_bool(value, f"{param_name}.{key}")
        elif key in RULE_ACTION_CHOICE_MAP:
            validated[key] = validate_choices(
                value,
                RULE_ACTION_CHOICE_MAP[key],
                f"{param_name}.{key}",
            )
    return validated


def _emailrules_list_impl(account_id: str) -> list[dict[str, Any]]:
    result = graph.request("GET", "/me/mailFolders/inbox/messageRules", account_id)
    if not result or "value" not in result:
        return []
    return list(result["value"])


def _emailrules_get_impl(rule_id: str, account_id: str) -> dict[str, Any]:
    result = graph.request(
        "GET", f"/me/mailFolders/inbox/messageRules/{rule_id}", account_id
    )
    if not result:
        raise ValueError(f"Message rule with ID {rule_id} not found")
    return result


# emailrules_list
@mcp.tool(
    name="emailrules_list",
    annotations={
        "title": "List Email Rules",
        "readOnlyHint": True,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={"category": "emailrules", "safety_level": "safe"},
)
def emailrules_list(account_id: str) -> list[dict[str, Any]]:
    """📖 List all inbox message rules (read-only, safe for unsupervised use)

    Message rules automatically process incoming emails based on conditions.
    Rules are executed in sequence order (1, 2, 3...).

    Args:
        account_id: Microsoft account ID

    Returns:
        List of rules with: id, displayName, sequence, isEnabled, conditions, actions
    """
    return _emailrules_list_impl(account_id)


# emailrules_get
@mcp.tool(
    name="emailrules_get",
    annotations={
        "title": "Get Email Rule",
        "readOnlyHint": True,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={"category": "emailrules", "safety_level": "safe"},
)
def emailrules_get(rule_id: str, account_id: str) -> dict[str, Any]:
    """📖 Get details of a specific message rule (read-only, safe for unsupervised use)

    Returns complete rule configuration including conditions, actions, and execution order.

    Args:
        rule_id: The message rule ID
        account_id: Microsoft account ID

    Returns:
        Rule details including conditions, actions, sequence, and enabled status
    """
    return _emailrules_get_impl(rule_id, account_id)


# emailrules_create
@mcp.tool(
    name="emailrules_create",
    annotations={
        "title": "Create Email Rule",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": False,
        "openWorldHint": True,
    },
    meta={"category": "emailrules", "safety_level": "moderate"},
)
def emailrules_create(
    account_id: str,
    display_name: str,
    conditions: dict[str, Any],
    actions: dict[str, Any],
    sequence: int = 1,
    is_enabled: bool = True,
    exceptions: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """✏️ Create a new inbox message rule to automatically process emails (requires user confirmation recommended)

    Rules are executed in priority order (sequence number).

    Conditions examples:
        {"fromAddresses": [{"address": "john@example.com"}]}
        {"subjectContains": ["urgent", "important"]}
        {"senderContains": ["@company.com"]}
        {"hasAttachments": true}

    Actions examples:
        {"moveToFolder": "folder_id"}
        {"markAsRead": true}
        {"forwardTo": [{"emailAddress": {"address": "manager@example.com"}}]}
        {"assignCategories": ["Red category"]}
        {"delete": true}

    Args:
        account_id: Microsoft account ID
        display_name: Name for the rule (e.g., "Move work emails to Projects")
        conditions: Conditions that trigger the rule
        actions: Actions to perform when conditions match
        sequence: Rule execution order (lower numbers execute first, default: 1)
        is_enabled: Whether the rule is active (default: True)
        exceptions: Optional conditions that prevent rule execution

    Returns:
        Created rule with its ID and full configuration
    """
    rule_data = {
        "displayName": display_name,
        "sequence": sequence,
        "isEnabled": is_enabled,
        "conditions": conditions,
        "actions": actions,
    }

    if exceptions:
        rule_data["exceptions"] = exceptions

    result = graph.request(
        "POST", "/me/mailFolders/inbox/messageRules", account_id, json=rule_data
    )
    if not result:
        raise ValueError("Failed to create message rule")
    return result


# emailrules_update
@mcp.tool(
    name="emailrules_update",
    annotations={
        "title": "Update Email Rule",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={"category": "emailrules", "safety_level": "moderate"},
)
def emailrules_update(
    rule_id: str,
    account_id: str,
    display_name: str | None = None,
    conditions: dict[str, Any] | None = None,
    actions: dict[str, Any] | None = None,
    sequence: int | None = None,
    is_enabled: bool | None = None,
    exceptions: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """✏️ Update an existing message rule (requires user confirmation recommended)

    Modifies rule properties, conditions, actions, or execution order.
    At least one field must be provided to update.

    Allowed parameters: display_name, conditions, actions, sequence,
    is_enabled, exceptions.

    Args:
        rule_id: The message rule ID to update
        account_id: Microsoft account ID
        display_name: New name for the rule (optional)
        conditions: New conditions (optional)
        actions: New actions (optional)
        sequence: New execution order (optional)
        is_enabled: Enable or disable the rule (optional)
        exceptions: New exception conditions (optional)

    Returns:
        Updated rule configuration
    """
    updates: dict[str, Any] = {}

    if display_name is not None:
        if not isinstance(display_name, str):
            raise ValidationError(
                format_validation_error(
                    "display_name",
                    display_name,
                    "must be a string",
                    "Rule display name",
                )
            )
        trimmed = display_name.strip()
        if not trimmed:
            raise ValidationError(
                format_validation_error(
                    "display_name",
                    display_name,
                    "cannot be empty",
                    "Non-empty rule name",
                )
            )
        updates["displayName"] = trimmed

    if conditions is not None:
        updates["conditions"] = _validate_rule_predicates(conditions, "conditions")

    if actions is not None:
        updates["actions"] = _validate_rule_actions(actions, "actions")

    if sequence is not None:
        if not isinstance(sequence, int):
            raise ValidationError(
                format_validation_error(
                    "sequence",
                    sequence,
                    "must be an integer",
                    "Integer value ≥ 1",
                )
            )
        if sequence < 1:
            raise ValidationError(
                format_validation_error(
                    "sequence",
                    sequence,
                    "must be at least 1",
                    "Integer value ≥ 1",
                )
            )
        updates["sequence"] = sequence

    if is_enabled is not None:
        updates["isEnabled"] = _validate_bool(is_enabled, "is_enabled")

    if exceptions is not None:
        updates["exceptions"] = _validate_rule_predicates(exceptions, "exceptions")

    if not updates:
        raise ValidationError(
            format_validation_error(
                "updates",
                {},
                "must include at least one update field",
                "Provide one or more allowed parameters",
            )
        )

    result = graph.request(
        "PATCH",
        f"/me/mailFolders/inbox/messageRules/{rule_id}",
        account_id,
        json=updates,
    )
    if not result:
        raise ValueError(f"Failed to update message rule {rule_id}")
    return result


# emailrules_delete
@mcp.tool(
    name="emailrules_delete",
    annotations={
        "title": "Delete Email Rule",
        "readOnlyHint": False,
        "destructiveHint": True,
        "idempotentHint": True,
        "openWorldHint": True,
    },
    meta={
        "category": "emailrules",
        "safety_level": "critical",
        "requires_confirmation": True,
    },
)
def emailrules_delete(
    rule_id: str, account_id: str, confirm: bool = False
) -> dict[str, str]:
    """🔴 Delete a message rule permanently (always require user confirmation)

    WARNING: This action permanently deletes the rule and cannot be undone.
    Emails will no longer be automatically processed by this rule.

    Args:
        rule_id: The message rule ID to delete
        account_id: Microsoft account ID
        confirm: Must be True to confirm deletion (prevents accidents)

    Returns:
        Status confirmation
    """
    require_confirm(confirm, "delete email rule")
    graph.request("DELETE", f"/me/mailFolders/inbox/messageRules/{rule_id}", account_id)
    return {"status": "deleted", "rule_id": rule_id}


# emailrules_move_top
@mcp.tool(
    name="emailrules_move_top",
    annotations={
        "title": "Move Email Rule to Top",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": False,
        "openWorldHint": True,
    },
    meta={"category": "emailrules", "safety_level": "moderate"},
)
def emailrules_move_top(rule_id: str, account_id: str) -> dict[str, Any]:
    """✏️ Move a message rule to the top of execution order (requires user confirmation recommended)

    Rules execute in sequence order. Moving to top means it runs before all other rules.
    Sets the rule's sequence number to 1.

    Args:
        rule_id: The message rule ID to move
        account_id: Microsoft account ID

    Returns:
        Updated rule with new sequence number
    """
    result = graph.request(
        "PATCH",
        f"/me/mailFolders/inbox/messageRules/{rule_id}",
        account_id,
        json={"sequence": 1},
    )
    if not result:
        raise ValueError(f"Failed to move rule {rule_id} to top")
    return result


# emailrules_move_bottom
@mcp.tool(
    name="emailrules_move_bottom",
    annotations={
        "title": "Move Email Rule to Bottom",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": False,
        "openWorldHint": True,
    },
    meta={"category": "emailrules", "safety_level": "moderate"},
)
def emailrules_move_bottom(rule_id: str, account_id: str) -> dict[str, Any]:
    """✏️ Move a message rule to the bottom of execution order (requires user confirmation recommended)

    Rules execute in sequence order. Moving to bottom means it runs after all other rules.

    Args:
        rule_id: The message rule ID to move
        account_id: Microsoft account ID

    Returns:
        Updated rule with new sequence number
    """
    # Get all rules to find the highest sequence number
    all_rules = _emailrules_list_impl(account_id)
    if not all_rules:
        raise ValueError("No rules found")

    max_sequence = max(rule.get("sequence", 1) for rule in all_rules)
    new_sequence = max_sequence + 1

    result = graph.request(
        "PATCH",
        f"/me/mailFolders/inbox/messageRules/{rule_id}",
        account_id,
        json={"sequence": new_sequence},
    )
    if not result:
        raise ValueError(f"Failed to move rule {rule_id} to bottom")
    return result


# emailrules_move_up
@mcp.tool(
    name="emailrules_move_up",
    annotations={
        "title": "Move Email Rule Up",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": False,
        "openWorldHint": True,
    },
    meta={"category": "emailrules", "safety_level": "moderate"},
)
def emailrules_move_up(rule_id: str, account_id: str) -> dict[str, Any]:
    """✏️ Move a message rule up one position in execution order (requires user confirmation recommended)

    Increases the rule's priority by moving it one position higher in the
    execution order. Rules at the top (sequence = 1) cannot be moved up.

    Args:
        rule_id: The message rule ID to move
        account_id: Microsoft account ID

    Returns:
        Updated rule with new sequence number
    """
    # Get current rule
    current_rule = _emailrules_get_impl(rule_id, account_id)
    current_sequence = current_rule.get("sequence", 1)

    if current_sequence <= 1:
        raise ValueError("Rule is already at the top (sequence = 1)")

    new_sequence = current_sequence - 1

    result = graph.request(
        "PATCH",
        f"/me/mailFolders/inbox/messageRules/{rule_id}",
        account_id,
        json={"sequence": new_sequence},
    )
    if not result:
        raise ValueError(f"Failed to move rule {rule_id} up")
    return result


# emailrules_move_down
@mcp.tool(
    name="emailrules_move_down",
    annotations={
        "title": "Move Email Rule Down",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": False,
        "openWorldHint": True,
    },
    meta={"category": "emailrules", "safety_level": "moderate"},
)
def emailrules_move_down(rule_id: str, account_id: str) -> dict[str, Any]:
    """✏️ Move a message rule down one position in execution order (requires user confirmation recommended)

    Decreases the rule's priority by moving it one position lower in the
    execution order. Rules at the bottom cannot be moved down further.

    Args:
        rule_id: The message rule ID to move
        account_id: Microsoft account ID

    Returns:
        Updated rule with new sequence number
    """
    # Get current rule
    current_rule = _emailrules_get_impl(rule_id, account_id)
    current_sequence = current_rule.get("sequence", 1)

    new_sequence = current_sequence + 1

    result = graph.request(
        "PATCH",
        f"/me/mailFolders/inbox/messageRules/{rule_id}",
        account_id,
        json={"sequence": new_sequence},
    )
    if not result:
        raise ValueError(f"Failed to move rule {rule_id} down")
    return result



from __future__ import annotations

from collections import deque
from pathlib import Path
from typing import Any, Callable

import httpx
import pytest

from src.microsoft_mcp.tools import file as file_tools
from src.microsoft_mcp.validators import ValidationError


@pytest.fixture
def record_stream_calls(monkeypatch: pytest.MonkeyPatch) -> Callable[[list[Any]], None]:
    """Patch _stream_download to replay configured outcomes."""

    calls: deque[Any] = deque()

    def configure(outcomes: list[Any]) -> None:
        calls.clear()
        calls.extend(outcomes)

    def fake_stream(url: str, destination: Path, **kwargs: Any) -> None:
        if not calls:
            raise AssertionError("No outcomes configured for _stream_download")
        outcome = calls.popleft()
        if isinstance(outcome, Exception):
            raise outcome
        data: bytes = outcome
        destination.write_bytes(data)

    monkeypatch.setattr(file_tools, "_stream_download", fake_stream)
    return configure


def _register_metadata(
    register_graph: Callable[[str, str, Any], None],
    metadata: dict[str, Any],
) -> None:
    register_graph("GET", f"/me/drive/items/{metadata['id']}", metadata)


def test_file_get_downloads_file_successfully(
    tmp_path: Path,
    mock_graph_request: Callable[[str, str, Any], None],
    mock_file_metadata: Callable[..., dict[str, Any]],
    mock_account_id: str,
    record_stream_calls: Callable[[list[Any]], None],
) -> None:
    destination = tmp_path / "downloaded.txt"
    metadata = mock_file_metadata(size=512)
    _register_metadata(mock_graph_request, metadata)
    record_stream_calls([b"hello world"])

    result = file_tools._file_get_impl(
        metadata["id"], mock_account_id, str(destination)
    )

    assert destination.exists()
    assert destination.read_bytes() == b"hello world"
    assert result["name"] == metadata["name"]
    assert result["mime_type"] == metadata["file"]["mimeType"]


def test_file_get_rejects_non_graph_host(
    tmp_path: Path,
    mock_graph_request: Callable[[str, str, Any], None],
    mock_file_metadata: Callable[..., dict[str, Any]],
    mock_account_id: str,
    record_stream_calls: Callable[[list[Any]], None],
) -> None:
    destination = tmp_path / "downloaded.txt"
    metadata = mock_file_metadata(
        **{"@microsoft.graph.downloadUrl": "https://malicious.example.com/file"}
    )
    _register_metadata(mock_graph_request, metadata)
    record_stream_calls([b""])

    with pytest.raises(ValidationError) as exc:
        file_tools._file_get_impl(metadata["id"], mock_account_id, str(destination))

    assert "Host" in exc.value.args[0] or "host" in exc.value.args[0]
    assert not destination.exists()


def test_file_get_enforces_size_limit(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    mock_graph_request: Callable[[str, str, Any], None],
    mock_file_metadata: Callable[..., dict[str, Any]],
    mock_account_id: str,
    record_stream_calls: Callable[[list[Any]], None],
) -> None:
    monkeypatch.setattr(file_tools, "MAX_DOWNLOAD_MIB", 1)
    destination = tmp_path / "downloaded.txt"
    metadata = mock_file_metadata(size=2 * 1024 * 1024)
    _register_metadata(mock_graph_request, metadata)
    record_stream_calls([b"a" * (2 * 1024)])

    with pytest.raises(ValidationError):
        file_tools._file_get_impl(metadata["id"], mock_account_id, str(destination))

    assert not destination.exists()


def test_file_get_cleans_up_on_http_error(
    tmp_path: Path,
    mock_graph_request: Callable[[str, str, Any], None],
    mock_file_metadata: Callable[..., dict[str, Any]],
    mock_account_id: str,
    record_stream_calls: Callable[[list[Any]], None],
) -> None:
    destination = tmp_path / "downloaded.txt"
    metadata = mock_file_metadata(size=512)
    _register_metadata(mock_graph_request, metadata)
    request = httpx.Request("GET", metadata["@microsoft.graph.downloadUrl"])
    response = httpx.Response(500, request=request)
    record_stream_calls(
        [
            httpx.HTTPStatusError("server error", request=request, response=response),
            httpx.HTTPStatusError("server error", request=request, response=response),
            httpx.HTTPStatusError("server error", request=request, response=response),
            httpx.HTTPStatusError("server error", request=request, response=response),
        ]
    )

    with pytest.raises(RuntimeError):
        file_tools._file_get_impl(metadata["id"], mock_account_id, str(destination))

    assert not destination.exists()


def test_file_get_retries_then_succeeds(
    tmp_path: Path,
    mock_graph_request: Callable[[str, str, Any], None],
    mock_file_metadata: Callable[..., dict[str, Any]],
    mock_account_id: str,
    record_stream_calls: Callable[[list[Any]], None],
) -> None:
    destination = tmp_path / "downloaded.txt"
    metadata = mock_file_metadata(size=512)
    _register_metadata(mock_graph_request, metadata)
    request = httpx.Request("GET", metadata["@microsoft.graph.downloadUrl"])
    response = httpx.Response(503, request=request)
    record_stream_calls(
        [
            httpx.HTTPStatusError("transient", request=request, response=response),
            b"final-chunk",
        ]
    )

    result = file_tools._file_get_impl(
        metadata["id"], mock_account_id, str(destination)
    )

    payload = destination.read_bytes()
    assert payload == b"final-chunk"
    expected_size_mb = round(len(payload) / (1024 * 1024), 2)
    assert result["size_mb"] == expected_size_mb


def test_file_get_timeout_raises_runtime_error(
    tmp_path: Path,
    mock_graph_request: Callable[[str, str, Any], None],
    mock_file_metadata: Callable[..., dict[str, Any]],
    mock_account_id: str,
    record_stream_calls: Callable[[list[Any]], None],
) -> None:
    destination = tmp_path / "downloaded.txt"
    metadata = mock_file_metadata(size=512)
    _register_metadata(mock_graph_request, metadata)
    record_stream_calls([httpx.ReadTimeout("timeout", request=None)])

    with pytest.raises(RuntimeError):
        file_tools._file_get_impl(metadata["id"], mock_account_id, str(destination))

    assert not destination.exists()


def test_file_list_rejects_invalid_limit(mock_account_id: str) -> None:
    with pytest.raises(ValidationError):
        file_tools.file_list.fn(account_id=mock_account_id, limit=0)



"""
Microsoft MCP Instance

This module contains the single FastMCP instance that all tools register with.
It's separated into its own module to avoid circular import issues.
"""

from fastmcp import FastMCP

# Create the single FastMCP instance that all tools will register with
mcp = FastMCP("microsoft-mcp")

__all__ = ["mcp"]



from __future__ import annotations

import datetime as dt
from typing import Any, Callable, Dict

import pytest

from src.microsoft_mcp.tools import search as search_tools
from src.microsoft_mcp.validators import ValidationError


@pytest.mark.parametrize(
    ("callable_fn", "kwargs"),
    [
        (
            search_tools.search_files.fn,
            {"query": "report", "account_id": "acc", "limit": 0},
        ),
        (
            search_tools.search_emails.fn,
            {"query": "report", "account_id": "acc", "limit": 0},
        ),
        (
            search_tools.search_contacts.fn,
            {"query": "user", "account_id": "acc", "limit": 0},
        ),
        (
            search_tools.search_unified.fn,
            {"query": "report", "account_id": "acc", "limit": 0},
        ),
    ],
)
def test_search_limits_reject_invalid_input(
    callable_fn: Callable[..., Any],
    kwargs: Dict[str, Any],
) -> None:
    with pytest.raises(ValidationError):
        callable_fn(**kwargs)


@pytest.mark.parametrize(
    ("callable_fn", "kwargs"),
    [
        (search_tools.search_files.fn, {"account_id": "acc"}),
        (search_tools.search_emails.fn, {"account_id": "acc"}),
        (search_tools.search_events.fn, {"account_id": "acc"}),
        (search_tools.search_contacts.fn, {"account_id": "acc"}),
        (search_tools.search_unified.fn, {"account_id": "acc"}),
    ],
)
@pytest.mark.parametrize("bad_query", ["", "   "])
def test_search_query_rejects_empty(
    callable_fn: Callable[..., Any],
    kwargs: Dict[str, Any],
    bad_query: str,
) -> None:
    with pytest.raises(ValidationError):
        callable_fn(query=bad_query, **kwargs)


@pytest.mark.parametrize(
    "callable_fn",
    [
        search_tools.search_files.fn,
        search_tools.search_emails.fn,
        search_tools.search_events.fn,
        search_tools.search_contacts.fn,
        search_tools.search_unified.fn,
    ],
)
def test_search_query_rejects_excess_length(callable_fn: Callable[..., Any]) -> None:
    overlong_query = "x" * 600
    with pytest.raises(ValidationError):
        callable_fn(query=overlong_query, account_id="acc")


def test_search_events_rejects_invalid_days(mock_account_id: str) -> None:
    with pytest.raises(ValidationError):
        search_tools.search_events.fn(
            query="meeting",
            account_id=mock_account_id,
            days_back=-1,
        )


def test_search_events_filters_by_range(
    monkeypatch: pytest.MonkeyPatch,
    mock_account_id: str,
) -> None:
    now = dt.datetime.now(dt.timezone.utc)
    inside_start = (now - dt.timedelta(hours=1)).isoformat()
    inside_end = (now + dt.timedelta(hours=1)).isoformat()
    outside_start = (now - dt.timedelta(days=10)).isoformat()
    outside_end = (now - dt.timedelta(days=9, hours=20)).isoformat()

    def fake_search_query(
        query: str,
        entity_types: list[str],
        account_id: str,
        limit: int,
    ) -> list[dict[str, Any]]:
        assert limit == 50
        return [
            {"start": {"dateTime": inside_start}, "end": {"dateTime": inside_end}},
            {"start": {"dateTime": outside_start}, "end": {"dateTime": outside_end}},
            {"start": {}, "end": {}},
        ]

    monkeypatch.setattr(search_tools.graph, "search_query", fake_search_query)

    results = search_tools.search_events.fn(
        query="meeting",
        account_id=mock_account_id,
        days_back=1,
        days_ahead=1,
        limit=50,
    )

    assert len(results) == 1
    assert results[0]["start"]["dateTime"] == inside_start


def test_search_emails_rejects_invalid_folder(mock_account_id: str) -> None:
    with pytest.raises(ValidationError):
        search_tools.search_emails.fn(
            query="reports",
            account_id=mock_account_id,
            folder="spam",
        )


def test_search_unified_rejects_invalid_entity_type() -> None:
    with pytest.raises(ValidationError):
        search_tools.search_unified.fn(
            query="report",
            account_id="acc",
            entity_types=["message", "calendar"],
        )


def test_search_files_trims_query(
    monkeypatch: pytest.MonkeyPatch,
    mock_account_id: str,
) -> None:
    captured: dict[str, Any] = {}

    def fake_search_query(
        query: str,
        entity_types: list[str],
        account_id: str,
        limit: int,
    ) -> list[dict[str, Any]]:
        captured["query"] = query
        captured["entity_types"] = entity_types
        captured["account_id"] = account_id
        captured["limit"] = limit
        return []

    monkeypatch.setattr(search_tools.graph, "search_query", fake_search_query)

    search_tools.search_files.fn(
        query="  quarterly report ",
        account_id=mock_account_id,
        limit=25,
    )

    assert captured["query"] == "quarterly report"
    assert captured["entity_types"] == ["driveItem"]
    assert captured["account_id"] == mock_account_id
    assert captured["limit"] == 25



from __future__ import annotations

from typing import Any

import pytest

from src.microsoft_mcp.tools import email_rules as email_rules_tools
from src.microsoft_mcp.validators import ValidationError


def test_emailrules_update_rejects_sequence_below_one(
    mock_account_id: str,
) -> None:
    with pytest.raises(ValidationError):
        email_rules_tools.emailrules_update.fn(
            rule_id="rule-1",
            account_id=mock_account_id,
            sequence=0,
        )


def test_emailrules_update_requires_updates(mock_account_id: str) -> None:
    with pytest.raises(ValidationError):
        email_rules_tools.emailrules_update.fn(
            rule_id="rule-1",
            account_id=mock_account_id,
        )


def test_emailrules_update_normalises_payload(
    monkeypatch: pytest.MonkeyPatch,
    mock_account_id: str,
) -> None:
    captured: dict[str, Any] = {}

    def fake_request(
        method: str,
        path: str,
        account_id: str | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        captured["json"] = kwargs.get("json") or {}
        return {"id": "rule-1"}

    monkeypatch.setattr(email_rules_tools.graph, "request", fake_request)

    result = email_rules_tools.emailrules_update.fn(
        rule_id="rule-1",
        account_id=mock_account_id,
        display_name="  Important Projects ",
        conditions={
            "categories": ["  Projects  "],
            "fromAddresses": ["LEAD@example.com"],
            "hasAttachments": True,
        },
        actions={
            "assignCategories": ["  Projects  "],
            "markAsRead": False,
            "forwardTo": ["manager@example.com"],
        },
    )

    assert result == {"id": "rule-1"}
    payload = captured["json"]
    assert payload["displayName"] == "Important Projects"
    assert payload["conditions"]["categories"] == ["Projects"]
    assert payload["conditions"]["fromAddresses"] == [
        {"emailAddress": {"address": "lead@example.com"}}
    ]
    assert payload["actions"]["assignCategories"] == ["Projects"]
    assert payload["actions"]["forwardTo"] == [
        {"emailAddress": {"address": "manager@example.com"}}
    ]


def test_emailrules_update_rejects_invalid_actions(mock_account_id: str) -> None:
    with pytest.raises(ValidationError):
        email_rules_tools.emailrules_update.fn(
            rule_id="rule-1",
            account_id=mock_account_id,
            actions={"assignCategories": "not-a-list"},
        )



#!/usr/bin/env python3
"""
Authenticate Microsoft accounts for use with M365 MCP.
Run this script to sign in to one or more Microsoft accounts.
"""

import os
import sys
import argparse
from pathlib import Path

# Add src to path so we can import our modules
sys.path.insert(0, str(Path(__file__).parent / "src"))

from dotenv import load_dotenv


def _parse_arguments() -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        description="Authenticate Microsoft accounts for M365 MCP"
    )
    parser.add_argument(
        "--env-file",
        type=Path,
        default=Path(".env"),
        help="Path to .env file (default: .env)",
    )
    return parser.parse_args()


def main():
    # Parse arguments first to get env file path
    args = _parse_arguments()

    # Load environment variables from custom path
    env_file = args.env_file
    if env_file.exists():
        load_dotenv(dotenv_path=env_file)
        print(f"Loaded environment from: {env_file}\n")
    else:
        print(f"Warning: Environment file not found: {env_file}")
        print("Continuing with system environment variables...\n")

    # Import auth module after loading environment
    from m365_mcp import auth
    if not os.getenv("M365_MCP_CLIENT_ID"):
        print("Error: M365_MCP_CLIENT_ID environment variable is required")
        print("\nPlease set it in your .env file or environment:")
        print("export M365_MCP_CLIENT_ID='your-app-id'")
        sys.exit(1)

    print("M365 MCP Authentication")
    print("============================\n")

    # List current accounts
    accounts = auth.list_accounts()
    if accounts:
        print("Currently authenticated accounts:")
        for i, account in enumerate(accounts, 1):
            print(f"{i}. {account.username} (ID: {account.account_id})")
        print()
    else:
        print("No accounts currently authenticated.\n")

    # Authenticate new account
    while True:
        choice = input("Do you want to authenticate a new account? (y/n): ").lower()
        if choice == "n":
            break
        elif choice == "y":
            try:
                # Use the new authentication function
                new_account = auth.authenticate_new_account()

                if new_account:
                    print("\n✓ Authentication successful!")
                    print(f"Signed in as: {new_account.username}")
                    print(f"Account ID: {new_account.account_id}")
                else:
                    print(
                        "\n✗ Authentication failed: Could not retrieve account information"
                    )
            except Exception as e:
                print(f"\n✗ Authentication failed: {e}")
                continue

            print()
        else:
            print("Please enter 'y' or 'n'")

    # Final account summary
    accounts = auth.list_accounts()
    if accounts:
        print("\nAuthenticated accounts summary:")
        print("==============================")
        for account in accounts:
            print(f"• {account.username}")
            print(f"  Account ID: {account.account_id}")

        print(
            "\nYou can use these account IDs with any MCP tool by passing account_id parameter."
        )
        print("Example: send_email(..., account_id='<account-id>')")
    else:
        print("\nNo accounts authenticated.")

    print("\nAuthentication complete!")


if __name__ == "__main__":
    main()



"""
Comprehensive logging configuration for Microsoft MCP Server.

This module provides structured logging with rotation, detailed formatting,
and separate log levels for different components.
"""

import logging
import logging.handlers
import os
import sys
import shutil
from pathlib import Path
from datetime import datetime, timezone
import json


class StructuredFormatter(logging.Formatter):
    """JSON structured formatter for machine-readable logs."""

    def format(self, record: logging.LogRecord) -> str:
        log_entry = {
            "timestamp": datetime.now(timezone.utc).isoformat(),
            "level": record.levelname,
            "logger": record.name,
            "message": record.getMessage(),
            "module": record.module,
            "function": record.funcName,
            "line": record.lineno,
        }

        # Add extra fields if present
        for key in ["account_id", "tool_name", "operation_id", "duration_ms"]:
            if hasattr(record, key):
                log_entry[key] = getattr(record, key)

        # Add exception info if present
        if record.exc_info:
            log_entry["exception"] = self.formatException(record.exc_info)

        return json.dumps(log_entry)


class HumanReadableFormatter(logging.Formatter):
    """Human-readable formatter with colors for console output."""

    COLORS = {
        "DEBUG": "\033[36m",  # Cyan
        "INFO": "\033[32m",  # Green
        "WARNING": "\033[33m",  # Yellow
        "ERROR": "\033[31m",  # Red
        "CRITICAL": "\033[35m",  # Magenta
        "RESET": "\033[0m",
    }

    def format(self, record: logging.LogRecord) -> str:
        # Add color if outputting to terminal
        if sys.stderr.isatty():
            color = self.COLORS.get(record.levelname, self.COLORS["RESET"])
            reset = self.COLORS["RESET"]
            record.levelname = f"{color}{record.levelname}{reset}"

        # Format: timestamp [LEVEL] logger.module.function:line - message
        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        location = f"{record.module}.{record.funcName}:{record.lineno}"

        formatted = f"{timestamp} [{record.levelname}] {record.name}.{location} - {record.getMessage()}"

        # Add exception info if present
        if record.exc_info:
            formatted += "\n" + self.formatException(record.exc_info)

        return formatted


def archive_existing_logs(log_dir: Path) -> dict[str, any]:
    """
    Archive existing log files to a timestamped folder.

    This ensures each server run gets fresh log files while preserving
    previous logs for historical reference.

    Args:
        log_dir: Directory containing log files to archive

    Returns:
        Dictionary with archival information:
        - 'archived': bool - Whether logs were archived
        - 'archive_dir': str or None - Archive directory path if archived
        - 'file_count': int - Number of files archived
    """
    result = {
        'archived': False,
        'archive_dir': None,
        'file_count': 0
    }

    if not log_dir.exists():
        return result

    # Check if any log files exist
    log_files = list(log_dir.glob("*.log*")) + list(log_dir.glob("*.jsonl*"))
    if not log_files:
        return result

    # Create archive directory with timestamp
    archive_timestamp = datetime.now(timezone.utc).strftime("%Y%m%d_%H%M%S")
    archive_dir = log_dir / "archives" / archive_timestamp
    archive_dir.mkdir(parents=True, exist_ok=True)

    # Move all log files to archive
    archived_count = 0
    for log_file in log_files:
        try:
            dest = archive_dir / log_file.name
            shutil.move(str(log_file), str(dest))
            archived_count += 1
        except Exception as e:
            # Print to stderr since logging isn't setup yet
            print(
                f"Warning: Failed to archive {log_file.name}: {e}",
                file=sys.stderr
            )

    result['archived'] = archived_count > 0
    result['archive_dir'] = str(archive_dir.relative_to(log_dir)) if archived_count > 0 else None
    result['file_count'] = archived_count

    return result


def setup_logging(
    log_dir: str = "logs",
    log_level: str = "INFO",
    max_bytes: int = 10 * 1024 * 1024,  # 10MB
    backup_count: int = 10,
) -> None:
    """
    Setup comprehensive logging for the MCP server.

    Existing log files are automatically archived to a timestamped folder
    under logs/archives/ on each startup, ensuring fresh logs for each run.

    Args:
        log_dir: Directory to store log files
        log_level: Minimum log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
        max_bytes: Maximum size of each log file before rotation
        backup_count: Number of backup log files to keep
    """
    # Create log directory
    log_path = Path(log_dir)
    log_path.mkdir(exist_ok=True)

    # Archive existing logs before starting new run
    archive_info = archive_existing_logs(log_path)

    # Convert log level string to logging constant
    numeric_level = getattr(logging, log_level.upper(), logging.INFO)

    # Get root logger
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)  # Capture all levels, filter per handler

    # Remove existing handlers to avoid duplicates
    root_logger.handlers.clear()

    # === File Handler: All logs (JSON structured) ===
    all_logs_file = log_path / "mcp_server_all.jsonl"
    all_handler = logging.handlers.RotatingFileHandler(
        all_logs_file,
        maxBytes=max_bytes,
        backupCount=backup_count,
        encoding="utf-8",
    )
    all_handler.setLevel(logging.DEBUG)
    all_handler.setFormatter(StructuredFormatter())
    root_logger.addHandler(all_handler)

    # === File Handler: Error logs only (JSON structured) ===
    error_logs_file = log_path / "mcp_server_errors.jsonl"
    error_handler = logging.handlers.RotatingFileHandler(
        error_logs_file,
        maxBytes=max_bytes,
        backupCount=backup_count,
        encoding="utf-8",
    )
    error_handler.setLevel(logging.ERROR)
    error_handler.setFormatter(StructuredFormatter())
    root_logger.addHandler(error_handler)

    # === File Handler: Human-readable logs ===
    readable_logs_file = log_path / "mcp_server.log"
    readable_handler = logging.handlers.RotatingFileHandler(
        readable_logs_file,
        maxBytes=max_bytes,
        backupCount=backup_count,
        encoding="utf-8",
    )
    readable_handler.setLevel(numeric_level)
    readable_handler.setFormatter(HumanReadableFormatter())
    root_logger.addHandler(readable_handler)

    # === Console Handler: Human-readable (stderr) ===
    console_handler = logging.StreamHandler(sys.stderr)
    console_handler.setLevel(numeric_level)
    console_handler.setFormatter(HumanReadableFormatter())
    root_logger.addHandler(console_handler)

    # Set specific log levels for noisy libraries
    logging.getLogger("httpx").setLevel(logging.WARNING)
    logging.getLogger("httpcore").setLevel(logging.WARNING)
    logging.getLogger("uvicorn").setLevel(logging.INFO)
    logging.getLogger("uvicorn.access").setLevel(logging.WARNING)

    # Log the startup
    logger = logging.getLogger("microsoft_mcp.logging")

    # Log archival status first
    if archive_info['archived']:
        logger.info(
            f"Previous logs archived: {archive_info['file_count']} file(s) → "
            f"{archive_info['archive_dir']}"
        )
    else:
        logger.info("Fresh start: No previous logs found")

    logger.info(f"Logging initialized - Level: {log_level}")
    logger.info(f"Log directory: {log_path.absolute()}")
    logger.info(f"All logs: {all_logs_file.name}")
    logger.info(f"Error logs: {error_logs_file.name}")
    logger.info(f"Readable logs: {readable_logs_file.name}")


def get_logger(name: str) -> logging.Logger:
    """
    Get a logger instance with the given name.

    Args:
        name: Logger name (typically __name__)

    Returns:
        Configured logger instance
    """
    return logging.getLogger(name)



import os
import asyncio
import json
from datetime import datetime, timedelta, timezone
from pathlib import Path
import pytest
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client
from dotenv import load_dotenv

# Support custom env file via TEST_ENV_FILE environment variable
test_env_file = os.getenv("TEST_ENV_FILE", ".env")
if Path(test_env_file).exists():
    load_dotenv(dotenv_path=test_env_file)
else:
    load_dotenv()

if not os.getenv("M365_MCP_CLIENT_ID"):
    pytest.fail("M365_MCP_CLIENT_ID environment variable is required")


def parse_result(result, tool_name=None):
    """Helper to parse MCP tool results consistently"""
    if result.content and hasattr(result.content[0], "text"):
        text = result.content[0].text
        if text == "[]":
            return []
        data = json.loads(text)
        # FastMCP seems to unwrap single-element lists, so rewrap for consistency
        list_tools = {
            "account_list",
            "email_list",
            "calendar_list_events",
            "contact_list",
            "file_list",
        }
        if tool_name in list_tools and isinstance(data, dict):
            return [data]
        return data
    return []


async def get_session():
    """Get MCP session"""
    server_params = StdioServerParameters(
        command="uv",
        args=["run", "m365-mcp"],
        env={
            "M365_MCP_CLIENT_ID": os.getenv("M365_MCP_CLIENT_ID", ""),
            "M365_MCP_TENANT_ID": os.getenv("M365_MCP_TENANT_ID", "common"),
        },
    )

    async with stdio_client(server_params) as (read, write):
        async with ClientSession(read, write) as session:
            await session.initialize()
            yield session


async def get_account_info(session):
    """Get account info"""
    result = await session.call_tool("account_list", {})
    assert not result.isError
    accounts = parse_result(result, "account_list")
    assert accounts and len(accounts) > 0, (
        "No accounts found - please authenticate first"
    )

    return {"email": accounts[0]["username"], "account_id": accounts[0]["account_id"]}


@pytest.mark.asyncio
async def test_list_accounts():
    """Test list_accounts tool"""
    async for session in get_session():
        result = await session.call_tool("account_list", {})
        assert not result.isError
        accounts = parse_result(result, "account_list")
        assert accounts is not None
        assert len(accounts) > 0
        assert "username" in accounts[0]
        assert "account_id" in accounts[0]


@pytest.mark.asyncio
async def test_list_emails():
    """Test list_emails tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        result = await session.call_tool(
            "email_list",
            {
                "account_id": account_info["account_id"],
                "limit": 3,
                "include_body": True,
            },
        )
        assert not result.isError
        emails = parse_result(result, "email_list")
        assert emails is not None
        if len(emails) > 0:
            assert "id" in emails[0]
            assert "subject" in emails[0]
            assert "body" in emails[0]


@pytest.mark.asyncio
async def test_list_emails_without_body():
    """Test list_emails tool without body"""
    async for session in get_session():
        account_info = await get_account_info(session)
        result = await session.call_tool(
            "email_list",
            {
                "account_id": account_info["account_id"],
                "limit": 3,
                "include_body": False,
            },
        )
        assert not result.isError
        emails = parse_result(result, "email_list")
        assert emails is not None
        if len(emails) > 0:
            assert "body" not in emails[0]


@pytest.mark.asyncio
async def test_get_email():
    """Test get_email tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        list_result = await session.call_tool(
            "email_list", {"account_id": account_info["account_id"], "limit": 1}
        )
        emails = parse_result(list_result, "email_list")

        if emails and len(emails) > 0:
            email_id = emails[0].get("id")
            result = await session.call_tool(
                "email_get",
                {"email_id": email_id, "account_id": account_info["account_id"]},
            )
            assert not result.isError
            email_detail = parse_result(result)
            assert email_detail is not None
            assert "id" in email_detail
            assert email_detail.get("id") == email_id


@pytest.mark.asyncio
async def test_create_email_draft():
    """Test create_email tool as draft"""
    async for session in get_session():
        account_info = await get_account_info(session)
        result = await session.call_tool(
            "email_create_draft",
            {
                "account_id": account_info["account_id"],
                "to": account_info["email"],
                "subject": "MCP Test Draft",
                "body": "This is a test draft email",
            },
        )
        assert not result.isError
        draft_data = parse_result(result)
        assert draft_data is not None
        assert "id" in draft_data

        draft_id = draft_data.get("id")
        delete_result = await session.call_tool(
            "email_delete",
            {
                "email_id": draft_id,
                "account_id": account_info["account_id"],
                "confirm": True,
            },
        )
        assert not delete_result.isError


@pytest.mark.asyncio
async def test_update_email():
    """Test update_email tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        list_result = await session.call_tool(
            "email_list", {"account_id": account_info["account_id"], "limit": 1}
        )
        emails = parse_result(list_result, "email_list")

        if emails and len(emails) > 0:
            email_id = emails[0].get("id")
            original_read_state = emails[0].get("isRead", True)

            result = await session.call_tool(
                "email_update",
                {
                    "email_id": email_id,
                    "account_id": account_info["account_id"],
                    "updates": {"isRead": not original_read_state},
                },
            )
            assert not result.isError

            restore_result = await session.call_tool(
                "email_update",
                {
                    "email_id": email_id,
                    "account_id": account_info["account_id"],
                    "updates": {"isRead": original_read_state},
                },
            )
            assert not restore_result.isError


@pytest.mark.asyncio
async def test_delete_email():
    """Test delete_email tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        draft_result = await session.call_tool(
            "email_create_draft",
            {
                "account_id": account_info["account_id"],
                "to": account_info["email"],
                "subject": "MCP Test Delete",
                "body": "This email will be deleted",
            },
        )
        draft_data = parse_result(draft_result)
        if draft_data and "id" in draft_data:
            result = await session.call_tool(
                "email_delete",
                {
                    "email_id": draft_data.get("id"),
                    "account_id": account_info["account_id"],
                    "confirm": True,
                },
            )
            assert not result.isError
            delete_result = parse_result(result)
            assert delete_result is not None
            assert delete_result.get("status") == "deleted"


@pytest.mark.asyncio
async def test_move_email():
    """Test move_email tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        list_result = await session.call_tool(
            "email_list",
            {"account_id": account_info["account_id"], "folder": "inbox", "limit": 1},
        )
        emails = parse_result(list_result, "email_list")

        if emails and len(emails) > 0:
            email_id = emails[0].get("id")
            result = await session.call_tool(
                "email_move",
                {
                    "email_id": email_id,
                    "account_id": account_info["account_id"],
                    "destination_folder": "archive",
                },
            )
            assert not result.isError

            move_result = parse_result(result, "email_move")
            new_email_id = move_result.get("new_id", email_id)

            restore_result = await session.call_tool(
                "email_move",
                {
                    "email_id": new_email_id,
                    "account_id": account_info["account_id"],
                    "destination_folder": "inbox",
                },
            )
            assert not restore_result.isError


@pytest.mark.asyncio
async def test_reply_to_email():
    """Test reply_to_email tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        await asyncio.sleep(2)
        list_result = await session.call_tool(
            "email_list", {"account_id": account_info["account_id"], "limit": 5}
        )
        emails = parse_result(list_result, "email_list")

        test_email = None
        if emails:
            test_email = next(
                (e for e in emails if "MCP Test" in e.get("subject", "")),
                emails[0] if emails else None,
            )

        if test_email:
            result = await session.call_tool(
                "email_reply",
                {
                    "account_id": account_info["account_id"],
                    "email_id": test_email.get("id"),
                    "body": "This is a test reply",
                    "confirm": True,
                },
            )
            assert not result.isError
            reply_result = parse_result(result)
            assert reply_result is not None
            assert reply_result.get("status") == "sent"


@pytest.mark.asyncio
async def test_reply_all_email():
    """Test reply_all_email tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        await asyncio.sleep(2)
        list_result = await session.call_tool(
            "email_list", {"account_id": account_info["account_id"], "limit": 5}
        )
        emails = parse_result(list_result, "email_list")

        test_email = None
        if emails:
            test_email = next(
                (e for e in emails if "MCP Test" in e.get("subject", "")),
                emails[0] if emails else None,
            )

        if test_email:
            result = await session.call_tool(
                "reply_all_email",
                {
                    "account_id": account_info["account_id"],
                    "email_id": test_email.get("id"),
                    "body": "This is a test reply to all",
                },
            )
            assert not result.isError
            reply_result = parse_result(result)
            assert reply_result is not None
            assert reply_result.get("status") == "sent"


@pytest.mark.asyncio
async def test_list_events():
    """Test list_events tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        result = await session.call_tool(
            "calendar_list_events",
            {
                "account_id": account_info["account_id"],
                "days_ahead": 14,
                "include_details": True,
            },
        )
        assert not result.isError
        events = parse_result(result, "calendar_list_events")
        assert events is not None
        if len(events) > 0:
            assert "id" in events[0]
            assert "subject" in events[0]
            assert "start" in events[0]
            assert "end" in events[0]


@pytest.mark.asyncio
async def test_get_event():
    """Test get_event tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        list_result = await session.call_tool(
            "calendar_list_events",
            {"account_id": account_info["account_id"], "days_ahead": 30},
        )
        events = parse_result(list_result, "calendar_list_events")

        if events and len(events) > 0:
            event_id = events[0].get("id")
            result = await session.call_tool(
                "calendar_get_event",
                {"event_id": event_id, "account_id": account_info["account_id"]},
            )
            assert not result.isError
            event_detail = parse_result(result)
            assert event_detail is not None
            assert "id" in event_detail
            assert event_detail.get("id") == event_id


@pytest.mark.asyncio
async def test_create_event():
    """Test create_event tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        start_time = datetime.now(timezone.utc) + timedelta(days=7)
        end_time = start_time + timedelta(hours=1)

        result = await session.call_tool(
            "calendar_create_event",
            {
                "account_id": account_info["account_id"],
                "subject": "MCP Integration Test Event",
                "start": start_time.isoformat(),
                "end": end_time.isoformat(),
                "location": "Virtual Meeting Room",
                "body": "This is a test event created by integration tests",
                "attendees": [account_info["email"]],
            },
        )
        assert not result.isError
        event_data = parse_result(result)
        assert event_data is not None
        assert "id" in event_data

        event_id = event_data.get("id")
        delete_result = await session.call_tool(
            "calendar_delete_event",
            {
                "account_id": account_info["account_id"],
                "event_id": event_id,
                "send_cancellation": False,
                "confirm": True,
            },
        )
        assert not delete_result.isError


@pytest.mark.asyncio
async def test_update_event():
    """Test update_event tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        start_time = datetime.now(timezone.utc) + timedelta(days=8)
        end_time = start_time + timedelta(hours=1)

        create_result = await session.call_tool(
            "calendar_create_event",
            {
                "account_id": account_info["account_id"],
                "subject": "MCP Test Event for Update",
                "start": start_time.isoformat(),
                "end": end_time.isoformat(),
            },
        )
        event_data = parse_result(create_result)
        assert event_data is not None
        event_id = event_data.get("id")

        new_start = start_time + timedelta(hours=2)
        new_end = new_start + timedelta(hours=1)

        result = await session.call_tool(
            "calendar_update_event",
            {
                "event_id": event_id,
                "account_id": account_info["account_id"],
                "updates": {
                    "subject": "MCP Test Event (Updated)",
                    "start": new_start.isoformat(),
                    "end": new_end.isoformat(),
                    "location": "Conference Room B",
                },
            },
        )
        assert not result.isError

        delete_result = await session.call_tool(
            "calendar_delete_event",
            {
                "account_id": account_info["account_id"],
                "event_id": event_id,
                "send_cancellation": False,
                "confirm": True,
            },
        )
        assert not delete_result.isError


@pytest.mark.asyncio
async def test_delete_event():
    """Test delete_event tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        start_time = datetime.now(timezone.utc) + timedelta(days=9)
        end_time = start_time + timedelta(hours=1)

        create_result = await session.call_tool(
            "calendar_create_event",
            {
                "account_id": account_info["account_id"],
                "subject": "MCP Test Event for Deletion",
                "start": start_time.isoformat(),
                "end": end_time.isoformat(),
            },
        )
        event_data = parse_result(create_result)
        assert event_data is not None
        event_id = event_data.get("id")

        result = await session.call_tool(
            "calendar_delete_event",
            {
                "account_id": account_info["account_id"],
                "event_id": event_id,
                "send_cancellation": False,
                "confirm": True,
            },
        )
        assert not result.isError
        delete_result = parse_result(result)
        assert delete_result is not None
        assert delete_result.get("status") == "deleted"


@pytest.mark.asyncio
async def test_respond_event():
    """Test respond_event tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        list_result = await session.call_tool(
            "calendar_list_events",
            {"account_id": account_info["account_id"], "days_ahead": 30},
        )
        events = parse_result(list_result, "calendar_list_events")

        if events:
            invite_event = next(
                (e for e in events if e.get("attendees") and len(e["attendees"]) > 1),
                None,
            )
            if invite_event:
                result = await session.call_tool(
                    "calendar_respond_event",
                    {
                        "account_id": account_info["account_id"],
                        "event_id": invite_event.get("id"),
                        "response": "tentativelyAccept",
                        "message": "I might be able to attend",
                    },
                )
                if not result.isError:
                    response_result = parse_result(result)
                    assert response_result is not None
                    assert response_result.get("status") == "tentativelyAccept"


@pytest.mark.asyncio
async def test_check_availability():
    """Test check_availability tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        check_start = (
            (datetime.now(timezone.utc) + timedelta(days=1))
            .replace(hour=10, minute=0)
            .isoformat()
        )
        check_end = (
            (datetime.now(timezone.utc) + timedelta(days=1))
            .replace(hour=17, minute=0)
            .isoformat()
        )

        result = await session.call_tool(
            "calendar_check_availability",
            {
                "account_id": account_info["account_id"],
                "start": check_start,
                "end": check_end,
                "attendees": [account_info["email"]],
            },
        )
        assert not result.isError
        availability = parse_result(result)
        assert availability is not None


@pytest.mark.asyncio
async def test_list_contacts():
    """Test list_contacts tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        result = await session.call_tool(
            "contact_list", {"account_id": account_info["account_id"], "limit": 10}
        )
        assert not result.isError
        contacts = parse_result(result, "contact_list")
        assert contacts is not None
        if len(contacts) > 0:
            assert "id" in contacts[0]


@pytest.mark.asyncio
async def test_get_contact():
    """Test get_contact tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        list_result = await session.call_tool(
            "contact_list", {"account_id": account_info["account_id"], "limit": 1}
        )
        assert not list_result.isError
        contacts = parse_result(list_result, "contact_list")
        if contacts and len(contacts) > 0:
            contact_id = contacts[0].get("id")
            result = await session.call_tool(
                "contact_get",
                {"contact_id": contact_id, "account_id": account_info["account_id"]},
            )
            assert not result.isError
            contact_detail = parse_result(result)
            assert contact_detail is not None
            assert "id" in contact_detail


@pytest.mark.asyncio
async def test_create_contact():
    """Test create_contact tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        result = await session.call_tool(
            "contact_create",
            {
                "account_id": account_info["account_id"],
                "given_name": "MCP",
                "surname": "TestContact",
                "email_addresses": ["mcp.test@example.com"],
                "phone_numbers": {"mobile": "+1234567890"},
            },
        )
        assert not result.isError
        new_contact = parse_result(result)
        assert new_contact is not None
        assert "id" in new_contact

        contact_id = new_contact.get("id")
        delete_result = await session.call_tool(
            "contact_delete",
            {
                "contact_id": contact_id,
                "account_id": account_info["account_id"],
                "confirm": True,
            },
        )
        assert not delete_result.isError


@pytest.mark.asyncio
async def test_update_contact():
    """Test update_contact tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        create_result = await session.call_tool(
            "contact_create",
            {
                "account_id": account_info["account_id"],
                "given_name": "MCPUpdate",
                "surname": "Test",
            },
        )
        assert not create_result.isError
        new_contact = parse_result(create_result)
        contact_id = new_contact.get("id")

        result = await session.call_tool(
            "contact_update",
            {
                "contact_id": contact_id,
                "account_id": account_info["account_id"],
                "updates": {"givenName": "MCPUpdated"},
            },
        )
        assert not result.isError

        delete_result = await session.call_tool(
            "contact_delete",
            {
                "contact_id": contact_id,
                "account_id": account_info["account_id"],
                "confirm": True,
            },
        )
        assert not delete_result.isError


@pytest.mark.asyncio
async def test_delete_contact():
    """Test delete_contact tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        create_result = await session.call_tool(
            "contact_create",
            {
                "account_id": account_info["account_id"],
                "given_name": "MCPDelete",
                "surname": "Test",
            },
        )
        assert not create_result.isError
        new_contact = parse_result(create_result)
        contact_id = new_contact.get("id")

        result = await session.call_tool(
            "contact_delete",
            {
                "contact_id": contact_id,
                "account_id": account_info["account_id"],
                "confirm": True,
            },
        )
        assert not result.isError
        delete_result = parse_result(result)
        assert delete_result is not None
        assert delete_result.get("status") == "deleted"


@pytest.mark.asyncio
async def test_list_files():
    """Test list_files tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        result = await session.call_tool(
            "file_list", {"account_id": account_info["account_id"]}
        )
        assert not result.isError
        files = parse_result(result)
        assert files is not None
        if len(files) > 0:
            assert "id" in files[0]
            assert "name" in files[0]
            assert "type" in files[0]


@pytest.mark.asyncio
async def test_get_file():
    """Test get_file tool"""
    import tempfile

    async for session in get_session():
        account_info = await get_account_info(session)
        test_content = "Test file content"
        test_filename = f"mcp-test-get-{datetime.now().strftime('%Y%m%d-%H%M%S')}.txt"

        # Create a temporary local file
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as local_file:
            local_file.write(test_content)
            local_file_path = local_file.name

        try:
            create_result = await session.call_tool(
                "file_create",
                {
                    "account_id": account_info["account_id"],
                    "onedrive_path": test_filename,
                    "local_file_path": local_file_path,
                },
            )
        finally:
            # Clean up local file
            if os.path.exists(local_file_path):
                os.unlink(local_file_path)
        file_data = parse_result(create_result)
        file_id = file_data.get("id")

        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            tmp_path = tmp_file.name

        try:
            result = await session.call_tool(
                "file_get",
                {
                    "file_id": file_id,
                    "account_id": account_info["account_id"],
                    "download_path": tmp_path,
                },
            )
            assert not result.isError
            retrieved_file = parse_result(result)
            assert retrieved_file is not None
            assert "path" in retrieved_file
            assert retrieved_file["path"] == tmp_path
            assert "name" in retrieved_file
            assert retrieved_file["name"] == test_filename
            assert "size_mb" in retrieved_file

            with open(tmp_path, "r") as f:
                downloaded_content = f.read()
            assert downloaded_content == test_content

        finally:
            if os.path.exists(tmp_path):
                os.unlink(tmp_path)

        delete_result = await session.call_tool(
            "file_delete",
            {
                "file_id": file_id,
                "account_id": account_info["account_id"],
                "confirm": True,
            },
        )
        assert not delete_result.isError


@pytest.mark.asyncio
async def test_create_file():
    """Test create_file tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        test_content = f"MCP Integration Test\nTimestamp: {datetime.now().isoformat()}"
        test_filename = (
            f"mcp-test-create-{datetime.now().strftime('%Y%m%d-%H%M%S')}.txt"
        )

        # Create a temporary local file
        import tempfile

        with tempfile.NamedTemporaryFile(mode="w", delete=False) as local_file:
            local_file.write(test_content)
            local_file_path = local_file.name

        try:
            result = await session.call_tool(
                "file_create",
                {
                    "account_id": account_info["account_id"],
                    "onedrive_path": test_filename,
                    "local_file_path": local_file_path,
                },
            )
        finally:
            # Clean up local file
            if os.path.exists(local_file_path):
                os.unlink(local_file_path)
        assert not result.isError
        upload_result = parse_result(result)
        assert upload_result is not None
        assert "id" in upload_result

        file_id = upload_result.get("id")
        delete_result = await session.call_tool(
            "file_delete",
            {
                "file_id": file_id,
                "account_id": account_info["account_id"],
                "confirm": True,
            },
        )
        assert not delete_result.isError


@pytest.mark.asyncio
async def test_update_file():
    """Test update_file tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        test_content = "Original content"
        test_filename = (
            f"mcp-test-update-{datetime.now().strftime('%Y%m%d-%H%M%S')}.txt"
        )

        # Create a temporary local file
        import tempfile

        with tempfile.NamedTemporaryFile(mode="w", delete=False) as local_file:
            local_file.write(test_content)
            local_file_path = local_file.name

        try:
            create_result = await session.call_tool(
                "file_create",
                {
                    "account_id": account_info["account_id"],
                    "onedrive_path": test_filename,
                    "local_file_path": local_file_path,
                },
            )
        finally:
            # Clean up local file
            if os.path.exists(local_file_path):
                os.unlink(local_file_path)
        file_data = parse_result(create_result)
        file_id = file_data.get("id")

        updated_content = f"Updated content at {datetime.now().isoformat()}"

        # Create a temporary local file with updated content
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as updated_file:
            updated_file.write(updated_content)
            updated_file_path = updated_file.name

        try:
            result = await session.call_tool(
                "file_update",
                {
                    "account_id": account_info["account_id"],
                    "file_id": file_id,
                    "local_file_path": updated_file_path,
                },
            )
        finally:
            # Clean up local file
            if os.path.exists(updated_file_path):
                os.unlink(updated_file_path)
        assert not result.isError

        delete_result = await session.call_tool(
            "file_delete",
            {
                "file_id": file_id,
                "account_id": account_info["account_id"],
                "confirm": True,
            },
        )
        assert not delete_result.isError


@pytest.mark.asyncio
async def test_delete_file():
    """Test delete_file tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        test_content = "File to be deleted"
        test_filename = (
            f"mcp-test-delete-{datetime.now().strftime('%Y%m%d-%H%M%S')}.txt"
        )

        # Create a temporary local file
        import tempfile

        with tempfile.NamedTemporaryFile(mode="w", delete=False) as local_file:
            local_file.write(test_content)
            local_file_path = local_file.name

        try:
            create_result = await session.call_tool(
                "file_create",
                {
                    "account_id": account_info["account_id"],
                    "onedrive_path": test_filename,
                    "local_file_path": local_file_path,
                },
            )
        finally:
            # Clean up local file
            if os.path.exists(local_file_path):
                os.unlink(local_file_path)
        file_data = parse_result(create_result)
        file_id = file_data.get("id")

        result = await session.call_tool(
            "file_delete",
            {
                "file_id": file_id,
                "account_id": account_info["account_id"],
                "confirm": True,
            },
        )
        assert not result.isError
        delete_result = parse_result(result)
        assert delete_result is not None
        assert delete_result.get("status") == "deleted"


@pytest.mark.asyncio
async def test_get_attachment():
    """Test get_attachment tool"""
    async for session in get_session():
        account_info = await get_account_info(session)

        # First create an email with an attachment
        import tempfile
        import os

        # Create a temporary directory and file with specific name
        temp_dir = tempfile.mkdtemp()
        temp_file_path = os.path.join(temp_dir, "test_file.txt")

        with open(temp_file_path, "w") as f:
            f.write("This is a test attachment content")

        try:
            draft_result = await session.call_tool(
                "email_create_draft",
                {
                    "account_id": account_info["account_id"],
                    "to": account_info["email"],
                    "subject": "MCP Test Email with Attachment",
                    "body": "This email contains a test attachment",
                    "attachments": temp_file_path,  # Test with single path
                },
            )
        finally:
            # Clean up temp file and directory
            if os.path.exists(temp_file_path):
                os.unlink(temp_file_path)
            if os.path.exists(temp_dir):
                os.rmdir(temp_dir)
        assert not draft_result.isError
        draft_data = parse_result(draft_result)
        email_id = draft_data["id"]

        # Get the email to retrieve attachment details
        email_result = await session.call_tool(
            "email_get",
            {
                "email_id": email_id,
                "account_id": account_info["account_id"],
            },
        )
        email_detail = parse_result(email_result)

        assert email_detail.get("attachments"), "Email should have attachments"
        attachment = email_detail["attachments"][0]

        # Test getting the attachment
        with tempfile.NamedTemporaryFile(suffix=".txt", delete=False) as save_file:
            save_path = save_file.name

        try:
            result = await session.call_tool(
                "email_get_attachment",
                {
                    "email_id": email_id,
                    "account_id": account_info["account_id"],
                    "attachment_id": attachment["id"],
                    "save_path": save_path,
                },
            )
            assert not result.isError
            attachment_data = parse_result(result)
            assert attachment_data is not None
            assert attachment_data["name"] == "test_file.txt"
            assert "saved_to" in attachment_data
            assert attachment_data["saved_to"] == save_path

            # Verify file was saved
            assert os.path.exists(save_path)
            with open(save_path, "r") as f:
                content = f.read()
                assert content == "This is a test attachment content"
        finally:
            # Clean up saved file
            if os.path.exists(save_path):
                os.unlink(save_path)

        # Clean up - delete the draft
        await session.call_tool(
            "email_delete",
            {
                "email_id": email_id,
                "account_id": account_info["account_id"],
                "confirm": True,
            },
        )


@pytest.mark.asyncio
async def test_search_files():
    """Test search_files tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        result = await session.call_tool(
            "search_files",
            {"account_id": account_info["account_id"], "query": "test", "limit": 5},
        )
        assert not result.isError
        search_results = parse_result(result)
        assert search_results is not None


@pytest.mark.asyncio
async def test_search_emails():
    """Test search_emails tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        result = await session.call_tool(
            "search_emails",
            {"account_id": account_info["account_id"], "query": "test", "limit": 5},
        )
        assert not result.isError
        search_results = parse_result(result)
        assert search_results is not None


@pytest.mark.asyncio
async def test_search_events():
    """Test search_events tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        result = await session.call_tool(
            "search_events",
            {"account_id": account_info["account_id"], "query": "meeting", "limit": 5},
        )
        assert not result.isError
        search_results = parse_result(result)
        assert search_results is not None


@pytest.mark.asyncio
async def test_search_contacts():
    """Test search_contacts tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        result = await session.call_tool(
            "search_contacts",
            {
                "account_id": account_info["account_id"],
                "query": account_info["email"].split("@")[0],
                "limit": 5,
            },
        )
        assert not result.isError
        search_results = parse_result(result)
        assert search_results is not None


@pytest.mark.asyncio
async def test_send_email():
    """Test send_email tool"""
    async for session in get_session():
        account_info = await get_account_info(session)
        await asyncio.sleep(2)

        result = await session.call_tool(
            "email_send",
            {
                "account_id": account_info["account_id"],
                "to": account_info["email"],
                "subject": f"MCP Test Send Email {datetime.now(timezone.utc).isoformat()}",
                "body": "This is a test email sent via send_email tool",
                "confirm": True,
            },
        )
        assert not result.isError
        sent_result = parse_result(result)
        assert sent_result is not None
        assert sent_result.get("status") == "sent"


@pytest.mark.asyncio
async def test_unified_search():
    """Test unified_search tool"""
    async for session in get_session():
        account_info = await get_account_info(session)

        result = await session.call_tool(
            "search_unified",
            {
                "account_id": account_info["account_id"],
                "query": "test",
                "entity_types": ["message"],
                "limit": 10,
            },
        )
        assert not result.isError
        search_results = parse_result(result)
        assert search_results is not None
        assert isinstance(search_results, dict)
        # Results should be grouped by entity type
        if "message" in search_results:
            assert isinstance(search_results["message"], list)
